#! /usr/bin/env python3
# -*- coding: utf-8 -*-

from ikabot.helpers.logging import getLogger
logger = getLogger(__name__)
import json
import os
import random
import re
import sys
import time
from requests import get
import ikabot.config as config
from ikabot.config import *
from ikabot.helpers.gui import *
from ikabot.helpers.pedirInfo import read


def sendToBotDebug(session, msg, debugON):
    """This function will send the ``msg`` argument passed to it as a message to the user on Telegram, only if ``debugOn`` is ``True``
    Parameters
    ----------
    session : ikabot.web.session.Session
        Session object
    msg : str
        a string representing the message to send to the user on Telegram
    debugON : bool
        a boolean indicating whether or not to send the message.
    """
    if debugON:
        sendToBot(session, msg)


def sendToBot(session, msg, Token=False, Photo=None):
    """This function will send the ``msg`` argument passed to it as a message to the user on Telegram
    Parameters
    ----------
    session : ikabot.web.session.Session
        Session object
    msg : str
        a string representing the message to send to the user on Telegram
    Token : bool
        a boolean indicating whether or not to attach the process id, the users server, world and Ikariam username to the message
    Photo : bytes
        a bytes object representing a picture to be sent.
    """

    logger.warning(f"MESSAGE TO TG BOT: {msg}", exc_info=True)

    if checkTelegramData(session) is False:
        logger.error("Tried to message TG bot without correct tg data!", exc_info=True)
        return
    if Token is False:
        # A bit hacky, but fixes issue that config.infoUser isn't updated in new processes for sending TG messages.
        infoUser = "Server:{}, World:{}, Player:{}".format(session.servidor, session.word, session.username)
        msg = "pid:{}\n{}\n{}".format(os.getpid(), infoUser, msg)

    sessionData = session.getSessionData()
    telegram_data = sessionData["shared"]["telegram"]
    if Photo is None:
        return get(
            "https://api.telegram.org/bot{}/sendMessage".format(
                telegram_data["botToken"]
            ),
            params={"chat_id": telegram_data["chatId"], "text": msg},
        )
    else:
        # we need to clear the headers here because telegram doesn't like keep-alive, might as well get rid of all headers
        headers = session.s.headers.copy()
        session.s.headers.clear()
        resp = session.s.post(
            "https://api.telegram.org/bot{}/sendDocument".format(
                telegram_data["botToken"]
            ),
            files={"document": ("captcha.png", Photo)},
            data={"chat_id": telegram_data["chatId"], "caption": msg},
        )
        session.s.headers = headers
        return resp


def telegramDataIsValid(session):
    """This function checks whether or not there is any Telegram data stored in the .ikabot file
    Parameters
    ----------
    session : ikabot.web.session.Session
        Session object

    Returns
    -------
    valid : bool
        a boolean indicating whether or not there is any Telegram data stored in the .ikabot file

    """
    sessionData = session.getSessionData()
    try:
        return (
            len(sessionData["shared"]["telegram"]["botToken"]) > 0
            and len(sessionData["shared"]["telegram"]["chatId"]) > 0
        )
    except KeyError:
        return False


def getUserResponse(session, fullResponse=False):
    """This function will retrieve a list of messages the user sent to the bot on Telegram.
    Parameters
    ----------
    session : ikabot.web.session.Session
        Session object

    Returns
    -------
    updates : list[str]
        a list containing all the messages the user sent to the bot on Telegram
    """
    # returns messages that the user sends to the telegram bot

    if checkTelegramData(session) is False:
        return []

    sessionData = session.getSessionData()
    telegram_data = sessionData["shared"]["telegram"]

    try:
        updates = get(
            "https://api.telegram.org/bot{}/getUpdates".format(
                telegram_data["botToken"]
            )
        ).text
        updates = json.loads(updates, strict=False)
        if updates["ok"] is False:
            return []
        updates = updates["result"]
        # only return messages from the chatId of our user
        if fullResponse:
            return [
                update["message"]
                for update in updates
                if "message" in update
                and update["message"]["chat"]["id"] == int(telegram_data["chatId"])
            ]
        else:
            return [
                update["message"]["text"]
                for update in updates
                if "message" in update
                and update["message"]["chat"]["id"] == int(telegram_data["chatId"])
            ]
    except KeyError:
        return []


def checkTelegramData(session):
    """This function doesn't actually check any data itself, that is done by the ``telegramDataIsValid`` function. This function returns ``True`` if there is any Telegram data in the .ikabot file, and if there is none, it will ask the user to input it.
    Parameters
    ----------
    session : ikabot.web.session.Session
        Session object

    Returns
    -------
    valid : bool
        a boolean indicating whether or not there is valid Telegram data in the .ikabot file.
    """
    if telegramDataIsValid(session):
        return True
    else:
        if not session.padre:  # stop asking people if process is detached
            return False
        banner()
        print("You must provide valid credentials to communicate by telegram.")
        print("You require the token of the bot you are going to use.")
        print("For more information about how to obtain them read the readme at https://github.com/Ikabot-Collective/ikabot"
        )
        rta = read(
            msg="Will you provide the credentials now? [y/N]",
            values=["y", "Y", "n", "N", ""],
        )
        if rta.lower() != "y":
            return False
        else:
            return updateTelegramData(session)


def updateTelegramData(session, event=None, stdin_fd=None, predetermined_input=[]):
    """This function asks the user to input the Telegram bot's token. After the user has input the token, this function will generate a random 4 digit number, and request of the user to send it as a command to their bot. Once the command has been sent to the bot, ikabot will save the incoming message's sender's chatid and save it into the session data.
    Parameters
    ----------
    session : ikabot.web.session.Session
        Session object
    event : multiprocessing.Event
        an event which, when fired, gives back control of the terminal to the main process
    stdin_fd : int
        the standard input file descriptor passed to the function as a means of gaining control of the terminal
    predetermined_input : multiprocessing.managers.SyncManager.list
        a process synced list of predetermined inputs

    Returns
    -------
    valid : bool
        a boolean indicating whether or not the Telegram data has been successfully updated
    """
    if event is not None and stdin_fd is not None:
        sys.stdin = os.fdopen(stdin_fd)  # give process access to terminal
    config.predetermined_input = predetermined_input
    banner()
    print("To create your own Telegram bot, read this: https://core.telegram.org/bots#3-how-do-i-create-a-bot")
    print("1. Just talk to @botfather in Telegram, send /newbot and then choose the bot's name.")
    print("2. Obtain your new bot's token")
    print("3. Remember to keep the token secret!\n")
    botToken = read(msg="Bot's token: ")

    updates = get(
        "https://api.telegram.org/bot{}/getUpdates".format(botToken)
    ).json()
    me = get(
        "https://api.telegram.org/bot{}/getMe".format(botToken)
    ).json()
    if "ok" not in updates or updates["ok"] is False:
        print("Invalid Telegram bot, try again.")
        enter()
        if event is not None and stdin_fd is not None:
            event.set()
        return False

    rand = str(random.randint(0, 9999)).zfill(4)
    print(f"\n{bcolors.GREEN}SUCCESS!{bcolors.ENDC} Telegram token is good!")
    print(
        f"\n4. Now send your bot the command {bcolors.BLUE}/ikabot {rand}{bcolors.ENDC} on Telegram.\nYour bot's username is @{me['result']['username']}"
    )

    start = time.time()
    user_id = None
    try:
        while True:
            print(
                f"Waiting to receive the command on Telegram... Press CTRL + C to abort.\tdt:{round(time.time()-start)}s",
                end="\r",
            )
            updates = get(
                "https://api.telegram.org/bot{}/getUpdates".format(botToken)
            ).json()

            for update in updates["result"]:
                if "message" in update:
                    if "text" in update["message"]:
                        if update["message"]["text"].strip() == f"/ikabot {rand}":
                            user_id = update["message"]["from"]["id"]
                            break
            time.sleep(2)
            print(" " * 100, end="\r")
            if user_id:
                break
    except KeyboardInterrupt:
        print(
            f"{bcolors.RED}FAILURE!{bcolors.ENDC} Did not find command {bcolors.BLUE}/ikabot {rand}{bcolors.ENDC} among received messages!\n\n{str(updates)}"
        )
        enter()
        if event is not None and stdin_fd is not None:
            event.set()
        return False

    telegram_data = {}
    telegram_data["telegram"] = {}
    telegram_data["telegram"]["botToken"] = botToken.replace(" ", "")
    telegram_data["telegram"]["chatId"] = str(user_id)
    session.setSessionData(telegram_data, shared=True)

    sendToBot(session, "You have successfully set up Telegram with ikabot.", Token=True)

    print(
        "\nA message was sent to you on Telegram informing you about the successful setup of the Telegram bot."
    )
    print(
        "If you did not receive any message on Telegram then something has gone wrong and you will need to set up the Telegram data again!"
    )
    enter()

    if event is not None and stdin_fd is not None:
        event.set()  # give main process control before exiting
    return True



#! /usr/bin/env python3
# -*- coding: utf-8 -*-

import json
import math
import re
import time
import traceback
from decimal import *

from ikabot.config import *
from ikabot.helpers.botComm import *
from ikabot.helpers.getJson import getCity
from ikabot.helpers.gui import *
from ikabot.helpers.naval import *
from ikabot.helpers.pedirInfo import *
from ikabot.helpers.planRoutes import waitForArrival
from ikabot.helpers.process import set_child_mode
from ikabot.helpers.signals import setInfoSignal
from ikabot.helpers.varios import *
from ikabot.helpers.pedirInfo import getShipCapacity

getcontext().prec = 30


def choose_island(session):
    idsIslands = getIslandsIds(session)
    islands = []
    for idIsland in idsIslands:
        html = session.get(island_url + idIsland)
        island = getIsland(html)
        islands.append(island)

    if len(islands) == 1:
        return islands[0]

    islands.sort(key=lambda island: island["id"])

    longest_island_name_length = 0
    for island in islands:
        longest_island_name_length = max(
            len(island["name"]), longest_island_name_length
        )

    def pad(island_name):
        return " " * (longest_island_name_length - len(island_name)) + island_name

    print("In which island do you want to attack the barbarians?")
    print(" 0) Exit")
    for i, island in enumerate(islands):
        num = " " + str(i + 1) if i < 9 else str(i + 1)
        if island["barbarians"]["destroyed"] == 1:
            warn = "(currently destroyed)"
        else:
            warn = ""
        print(
            "{}) [{}:{}] {} ({}) : barbarians lv: {} ({}) {}".format(
                num,
                island["x"],
                island["y"],
                pad(island["name"]),
                materials_names[int(island["tradegood"])][0].upper(),
                island["barbarians"]["level"],
                island["barbarians"]["city"],
                warn,
            )
        )

    index = read(min=0, max=len(islands))
    if index == 0:
        return None
    else:
        return islands[index - 1]


def get_barbarians_lv(session, island, ship_capacity):
    params = {
        "view": "barbarianVillage",
        "destinationIslandId": island["id"],
        "oldBackgroundView": "city",
        "cityWorldviewScale": "1",
        "islandId": island["id"],
        "backgroundView": "island",
        "currentIslandId": island["id"],
        "actionRequest": actionRequest,
        "ajax": "1",
    }
    resp = session.post(params=params)
    resp = json.loads(resp, strict=False)

    level = int(math.ceil(resp[2][1]["js_islandBarbarianLevel"]["text"]))
    gold = int(resp[2][1]["js_islandBarbarianResourcegold"]["text"].replace(",", ""))

    resources = [0] * len(materials_names)
    for i in range(len(materials_names)):
        if i == 0:
            resources[i] = int(
                resp[2][1]["js_islandBarbarianResourceresource"]["text"].replace(
                    ",", ""
                )
            )
        else:
            resources[i] = int(
                resp[2][1]["js_islandBarbarianResourcetradegood{:d}".format(i)][
                    "text"
                ].replace(",", "")
            )

    html = resp[1][1][1]
    troops = re.findall(
        r'<div class="army \w*?">\s*<div class=".*?">(.*?)</div>\s*</div>\s*</td>\s*</tr>\s*<tr>\s*<td class="center">\s*(\d+)',
        html,
    )

    total_cargo = sum(resources)
    ships = math.ceil(Decimal(total_cargo) / Decimal(ship_capacity))

    info = {
        "island_id": island["id"],
        "level": level,
        "gold": gold,
        "resources": resources,
        "troops": troops,
        "ships": ships,
    }

    return info


def get_units(session, city):
    params = {
        "view": "cityMilitary",
        "activeTab": "tabUnits",
        "cityId": city["id"],
        "backgroundView": "city",
        "currentCityId": city["id"],
        "currentTab": "multiTab1",
        "actionRequest": actionRequest,
        "ajax": "1",
    }

    resp = session.post(params=params)
    resp = json.loads(resp, strict=False)
    html = resp[1][1][1]
    html = html.split('<div class="fleet')[0]

    unit_id_names = re.findall(
        r'<div class="army (.*?)">\s*<div class="tooltip">(.*?)<\/div>', html
    )
    unit_amounts = re.findall(r"<td>(.*?)\s*</td>", html)

    units = {}
    for i in range(len(unit_id_names)):
        amount = int(unit_amounts[i].replace(",", "").replace("-", "0"))
        unit_id = unit_id_names[i][0][1:]
        unit_name = unit_id_names[i][1]
        units[unit_id] = {}
        units[unit_id]["name"] = unit_name
        units[unit_id]["amount"] = amount

    return units


def plan_attack(session, city, babarians_info):
    total_units = get_units(session, city)

    if sum([total_units[unit_id]["amount"] for unit_id in total_units]) == 0:
        print("You don't have any troops in this city!")
        enter()
        return None

    plan = []
    total_ships = None
    last = False
    while True:

        banner()

        units_available = {}
        for unit_id in total_units:

            already_sent = sum(
                [p["units"][u] for p in plan for u in p["units"] if u == unit_id]
            )
            if already_sent < total_units[unit_id]["amount"]:
                units_available[unit_id] = {}
                units_available[unit_id]["amount"] = (
                    total_units[unit_id]["amount"] - already_sent
                )
                units_available[unit_id]["name"] = total_units[unit_id]["name"]

        if len(units_available) == 0:
            print("No more troops available to send")
            enter()
            break

        attack_round = {}
        attack_round["units"] = {}
        print("Which troops do you want to send?")
        for unit_id in units_available:
            unit_amount = units_available[unit_id]["amount"]
            unit_name = units_available[unit_id]["name"]
            amount_to_send = read(
                msg="{} (max: {}): ".format(
                    unit_name, addThousandSeparator(unit_amount)
                ),
                max=unit_amount,
                default=0,
            )
            if amount_to_send > 0:
                attack_round["units"][unit_id] = amount_to_send
        print("")

        attack_round["loot"] = last
        if last:
            attack_round["round"] = len(plan) + 1
        else:
            if len(plan) > 0:
                round_def = len(plan) + 1
                attack_round["round"] = read(
                    msg= "In which battle round do you want to send them? (min: 2, default: {:d}): ".format(round_def),
                    min=2,
                    default=round_def,
                )
            else:
                attack_round["round"] = 1
        print("")

        if last is False:
            if total_ships is None:
                total_ships = getTotalShips(session)
            max_ships = total_ships - sum([ar["ships"] for ar in plan])
            if max_ships > 0:
                attack_round["ships"] = read(
                    msg= "How many ships do you want to send in this round? (min: 0, max: {:d}): ".format(max_ships),
                    min=0,
                    max=max_ships,
                )
                print("")
            else:
                attack_round["ships"] = 0

        plan.append(attack_round)

        if last:
            break

        print("Do you want to send another round of troops? [y/N]")
        resp = read(values=["y", "Y", "n", "N"], default="n")
        if resp.lower() != "y":
            print("")
            print("Do you want to select the troops that will be used to collect the remaining resources? (they need to destroy the wall) [y/N]")
            resp = read(values=["y", "Y", "n", "N"], default="n")
            if resp.lower() != "y":
                break
            else:
                last = True

    plan.sort(key=lambda ar: ar["round"])
    return plan


def attackBarbarians(session, event, stdin_fd, predetermined_input):
    """
    Parameters
    ----------
    session : ikabot.web.session.Session
    event : multiprocessing.Event
    stdin_fd: int
    predetermined_input : multiprocessing.managers.SyncManager.list
    """
    sys.stdin = os.fdopen(stdin_fd)
    config.predetermined_input = predetermined_input
    try:
        banner()

        island = choose_island(session)
        if island is None:
            event.set()
            return
        ship_capacity, freighter_capacity = getShipCapacity(session)
        babarians_info = get_barbarians_lv(session, island, ship_capacity)

        banner()
        print("The barbarians have:")
        for name, amount in babarians_info["troops"]:
            print("{} units of {}".format(amount, name))
        print("")

        banner()
        print("From which city do you want to attack?")
        city = chooseCity(session)

        plan = plan_attack(session, city, babarians_info)
        if plan is None:
            event.set()
            return

        banner()
        print(
            "The barbarians in [{}:{}] will be attacked.".format(
                island["x"], island["y"]
            )
        )
        enter()

    except KeyboardInterrupt:
        event.set()
        return

    set_child_mode(session)
    event.set()

    info = "\nI attack the barbarians in [{}:{}]\n".format(island["x"], island["y"])
    setInfoSignal(session, info)
    try:
        do_it(session, island, city, babarians_info, plan, ship_capacity)
    except Exception as e:
        msg = "Error in:\n{}\nCause:\n{}".format(info, traceback.format_exc())
        sendToBot(session, msg)
    finally:
        session.logout()


def get_unit_data(session, city_id, unit_id):
    params_w = {
        "view": "unitdescription",
        "unitId": unit_id,
        "helpId": 9,
        "subHelpId": 0,
        "backgroundView": "city",
        "currentCityId": city_id,
        "templateView": "unitdescription",
        "actionRequest": actionRequest,
        "ajax": 1,
    }
    resp = session.post(params=params_w)
    resp = json.loads(resp, strict=False)
    html = resp[1][1][1]

    weight = re.search(
        r'<li class="weight fifthpos" title=".*?"><span\s*class="accesshint">\'.*?\': </span>(\d+)</li>',
        html,
    ).group(1)
    weight = int(weight)

    speed = re.search(
        r'<span class="textLabel">.*?</span>\s*(\d+)\s*<br/>',
        html,
    ).group(1)
    speed = int(speed)

    name = re.search(r'<div class="contentBox01h">\s*<h3 class="header">([^<]+)</h3>', html).group(1)
    name = str(name)

    return {"speed": speed, "weight": weight, "name": name}


def city_is_in_island(city, island):
    return city["x"] == island["x"] and city["y"] == island["y"]


def get_barbarians_info(session, island_id):
    query = {
        "view": "barbarianVillage",
        "destinationIslandId": island_id,
        "backgroundView": "island",
        "currentIslandId": island_id,
        "actionRequest": actionRequest,
        "ajax": 1,
    }
    resp = session.post(params=query)
    resp = json.loads(resp, strict=False)
    return resp


def wait_until_can_attack(session, city, island, travel_time=0):
    html = session.get(island_url + island["id"])
    island = getIsland(html)

    if (
        island["barbarians"]["underAttack"] == 0
        and island["barbarians"]["destroyed"] == 0
    ):
        # an attack might be on its way
        wait_for_arrival(session, city, island)
        html = session.get(island_url + island["id"])
        island = getIsland(html)

    if island["barbarians"]["underAttack"] == 1:
        # a battle is taking place
        attacks = get_current_attacks(session, city["id"], island["id"])
        attacks_fighting = filter_fighting(attacks)
        eventTimes = [attack["eventTime"] for attack in attacks_fighting]
        if len(eventTimes) > 0:
            wait_time = max(eventTimes)
            wait_time -= time.time()
            wait(wait_time + 5)
        wait_until_can_attack(session, city, island, travel_time)

    if island["barbarians"]["destroyed"] == 1:
        # the barbarians are destroyed and can't be attacked
        resp = get_barbarians_info(session, island["id"])
        if "barbarianCityCooldownTimer" in resp[2][1]:
            wait_time = resp[2][1]["barbarianCityCooldownTimer"]["countdown"]["enddate"]
            wait_time -= time.time()
            wait_time -= travel_time
            wait(wait_time + 5)
        wait_until_can_attack(session, city, island, travel_time)


def get_movements(session, city_id=None):
    if city_id is None:
        city_id = getCurrentCityId(session)
    query = {
        "view": "militaryAdvisor",
        "oldView": "updateGlobalData",
        "cityId": city_id,
        "backgroundView": "city",
        "currentCityId": city_id,
        "templateView": "militaryAdvisor",
        "actionRequest": actionRequest,
        "ajax": 1,
    }

    resp = session.post(params=query)
    resp = json.loads(resp, strict=False)
    movements = resp[1][1][2]["viewScriptParams"]["militaryAndFleetMovements"]

    return movements


def get_current_attacks(session, city_id, island_id):

    movements = get_movements(session, city_id)
    curr_attacks = []

    for movement in movements:
        if movement["event"]["mission"] != 13:
            continue
        if movement["target"]["islandId"] != int(island_id):
            continue
        if movement["event"]["isReturning"] != 0:
            continue
        if movement["origin"]["cityId"] == -1:
            continue

        curr_attacks.append(movement)

    return curr_attacks


def wait_for_arrival(session, city, island):
    attacks = get_current_attacks(session, city["id"], island["id"])
    attacks = filter_loading(attacks) + filter_traveling(attacks)
    eventTimes = [attack["eventTime"] for attack in attacks]

    if len(eventTimes) == 0:
        return

    wait_time = max(eventTimes)
    wait_time -= time.time()
    wait(wait_time + 5)

    wait_for_arrival(session, city, island)


def wait_for_round(session, city, island, travel_time, battle_start, round_number):
    if round_number == 1:
        wait_until_can_attack(session, city, island, travel_time)
    else:
        wait_time = battle_start + (round_number - 2) * 15 * 60
        wait_time -= time.time()
        wait_time -= travel_time
        wait(wait_time + 5)

        if battle_start < time.time():
            html = session.get(island_url + city["id"])
            island = getIsland(html)
            assert (
                island["barbarians"]["underAttack"] == 1
            ), "the battle ended before expected"


def calc_travel_time(city, island, speed):
    if city["x"] == island["x"] and city["y"] == island["y"]:
        return math.ceil(36000 / speed)
    else:
        return math.ceil(
            1200
            * math.sqrt(
                ((city["x"] - island["x"]) ** 2) + ((city["y"] - island["y"]) ** 2)
            )
        )


def filter_loading(attacks):
    return [attack for attack in attacks if attack["event"]["missionState"] == 1]


def filter_traveling(attacks,onlyCanAbort=True):
    return [
        attack
        for attack in attacks
        if attack["event"]["missionState"] == 2 and (onlyCanAbort is False or (onlyCanAbort is True and attack["event"]["canAbort"]))
    ]


def filter_fighting(attacks):
    return [
        attack
        for attack in attacks
        if attack["event"]["missionState"] == 2 and attack["event"]["canRetreat"]
    ]


def wait_until_attack_is_over(session, city, island):
    html = session.get(island_url + island["id"])
    island = getIsland(html)

    while island["barbarians"]["destroyed"] == 0:

        attacks = get_current_attacks(session, city["id"], island["id"])
        # the attack probably failed
        if len(attacks) == 0:
            return

        eventTimes = [attack["eventTime"] for attack in attacks]
        wait_time = min(eventTimes)
        wait_time -= time.time()
        wait(wait_time + 5)

        html = session.get(island_url + island["id"])
        island = getIsland(html)


def load_troops(
    session, city, island, attack_round, units_data, attack_data, ship_capacity, extra_cargo=0
):
    ships_needed = Decimal(extra_cargo) / Decimal(ship_capacity)
    speeds = []
    current_units = get_units(session, city)
    for unit_id in attack_round["units"]:
        amount_to_send = min(
            attack_round["units"][unit_id], current_units[unit_id]["amount"]
        )
        attack_data["cargo_army_{}".format(unit_id)] = amount_to_send

        if unit_id not in units_data:
            units_data[unit_id] = get_unit_data(session, city["id"], unit_id)

        speeds.append(units_data[unit_id]["speed"])

        if city_is_in_island(city, island) is False:
            weight = units_data[unit_id]["weight"]
            ships_needed += Decimal(amount_to_send * weight) / Decimal(ship_capacity)

    ships_needed = math.ceil(ships_needed)
    speed = min(speeds)
    travel_time = calc_travel_time(city, island, speed)
    return attack_data, ships_needed, travel_time


def loot(session, island, city, units_data, loot_round, ship_capacity):
    while True:

        attack_data = {
            "action": "transportOperations",
            "function": "attackBarbarianVillage",
            "actionRequest": actionRequest,
            "islandId": island["id"],
            "destinationCityId": 0,
            "cargo_army_304_upkeep": 3,
            "cargo_army_304": 0,
            "cargo_army_315_upkeep": 1,
            "cargo_army_315": 0,
            "cargo_army_302_upkeep": 4,
            "cargo_army_302": 0,
            "cargo_army_303_upkeep": 3,
            "cargo_army_303": 0,
            "cargo_army_312_upkeep": 15,
            "cargo_army_312": 0,
            "cargo_army_309_upkeep": 45,
            "cargo_army_309": 0,
            "cargo_army_307_upkeep": 15,
            "cargo_army_307": 0,
            "cargo_army_306_upkeep": 25,
            "cargo_army_306": 0,
            "cargo_army_305_upkeep": 30,
            "cargo_army_305": 0,
            "cargo_army_311_upkeep": 20,
            "cargo_army_311": 0,
            "cargo_army_310_upkeep": 10,
            "cargo_army_310": 0,
            "transporter": 0,
            "barbarianVillage": 1,
            "backgroundView": "island",
            "currentIslandId": island["id"],
            "templateView": "plunder",
            "ajax": 1,
        }

        # make sure we have ships on the port
        ships_available = waitForArrival(session)

        # if the barbarians are active again or all the resources were stolen, return
        html = session.get(island_url + island["id"])
        island = getIsland(html)
        destroyed = island["barbarians"]["destroyed"] == 1
        resources = get_barbarians_lv(session, island, ship_capacity)["resources"]
        if destroyed is False or sum(resources) == 0:
            return

        # if we already sent an attack and we still have ships on the port, it was the last one
        attacks = get_current_attacks(session, city["id"], island["id"])
        attacks = filter_loading(attacks) + filter_traveling(attacks)
        if len(attacks) > 0:
            return

        attack_data, ships_needed, travel_time = load_troops(
            session, city, island, loot_round, units_data, attack_data, ship_capacity, sum(resources)
        )
        attack_data["transporter"] = min(ships_available, ships_needed)

        # make sure we have time to send the attack
        time_left = None
        resp = get_barbarians_info(session, island["id"])
        if "barbarianCityCooldownTimer" in resp[2][1]:
            time_left = resp[2][1]["barbarianCityCooldownTimer"]["countdown"]["enddate"]
            time_left -= time.time()
        if time_left is not None and travel_time > time_left:
            return

        # send attack
        session.post(params=attack_data)


def do_it(session, island, city, babarians_info, plan, ship_capacity):

    units_data = {}

    battle_start = None

    for attack_round in plan:

        # this round is supposed to get the resources
        if attack_round["loot"]:
            break

        attack_data = {
            "action": "transportOperations",
            "function": "attackBarbarianVillage",
            "actionRequest": actionRequest,
            "islandId": island["id"],
            "destinationCityId": 0,
            "cargo_army_304_upkeep": 3,
            "cargo_army_304": 0,
            "cargo_army_315_upkeep": 1,
            "cargo_army_315": 0,
            "cargo_army_302_upkeep": 4,
            "cargo_army_302": 0,
            "cargo_army_303_upkeep": 3,
            "cargo_army_303": 0,
            "cargo_army_312_upkeep": 15,
            "cargo_army_312": 0,
            "cargo_army_309_upkeep": 45,
            "cargo_army_309": 0,
            "cargo_army_307_upkeep": 15,
            "cargo_army_307": 0,
            "cargo_army_306_upkeep": 25,
            "cargo_army_306": 0,
            "cargo_army_305_upkeep": 30,
            "cargo_army_305": 0,
            "cargo_army_311_upkeep": 20,
            "cargo_army_311": 0,
            "cargo_army_310_upkeep": 10,
            "cargo_army_310": 0,
            "transporter": 0,
            "barbarianVillage": 1,
            "backgroundView": "island",
            "currentIslandId": island["id"],
            "templateView": "plunder",
            "ajax": 1,
        }

        attack_data, ships_needed, travel_time = load_troops(
            session, city, island, attack_round, units_data, attack_data, ship_capacity
        )

        try:
            wait_for_round(
                session, city, island, travel_time, battle_start, attack_round["round"]
            )
        except AssertionError:
            # battle ended before expected
            break

        ships_available = waitForArrival(session)
        while ships_available < ships_needed:
            ships_available = waitForArrival(session)
        ships_available -= ships_needed

        # if the number of available troops changed, the POST request might not work as intended

        attack_data["transporter"] = min(
            babarians_info["ships"], attack_round["ships"], ships_available
        )

        # send new round
        session.post(params=attack_data)

        if attack_round["round"] == 1:
            battle_start = time.time() + travel_time

    wait_until_attack_is_over(session, city, island)

    last_round = plan[-1]
    if last_round["loot"]:
        loot(session, city, island, units_data, last_round, ship_capacity)



#! /usr/bin/env python3
# -*- coding: utf-8 -*-

import datetime
import multiprocessing
import os
import sys
import time

from ikabot.config import *
from ikabot.function.activateMiracle import activateMiracle
from ikabot.function.alertAttacks import alertAttacks
from ikabot.function.alertLowWine import alertLowWine
from ikabot.function.attackBarbarians import attackBarbarians
from ikabot.function.autoBarbarians import autoBarbarians
from ikabot.function.autoPirate import autoPirate
from ikabot.function.buyResources import buyResources
from ikabot.function.checkForUpdate import checkForUpdate
from ikabot.function.constructBuilding import constructBuilding
from ikabot.function.constructionList import constructionList
from ikabot.function.decaptchaConf import decaptchaConf
from ikabot.function.distributeResources import distributeResources
from ikabot.function.donate import donate
from ikabot.function.donationBot import donationBot
from ikabot.function.dumpWorld import dumpWorld
from ikabot.function.getStatus import getStatus
from ikabot.function.importExportCookie import importExportCookie
from ikabot.function.investigate import investigate
from ikabot.function.consolidateResources import consolidateResources
from ikabot.function.killTasks import killTasks
from ikabot.function.loginDaily import loginDaily
from ikabot.function.logs import logs
from ikabot.function.proxyConf import proxyConf, show_proxy
from ikabot.function.searchForIslandSpaces import searchForIslandSpaces
from ikabot.function.sellResources import sellResources
from ikabot.function.sendResources import sendResources
from ikabot.function.shipMovements import shipMovements
from ikabot.function.stationArmy import stationArmy
from ikabot.function.testTelegramBot import testTelegramBot
from ikabot.function.trainArmy import trainArmy
from ikabot.function.update import update
from ikabot.function.vacationMode import vacationMode
from ikabot.function.webServer import webServer
from ikabot.function.loadCustomModule import loadCustomModule
from ikabot.function.activateShrine import activateShrine
from ikabot.helpers.botComm import telegramDataIsValid, updateTelegramData
from ikabot.helpers.gui import *
from ikabot.helpers.pedirInfo import read
from ikabot.helpers.process import updateProcessList
from ikabot.web.session import *
from ikabot.function.modifyProduction import modifyProduction


def menu(session, checkUpdate=True):
    """
    Parameters
    ----------
    session : ikabot.web.session.Session
    checkUpdate : bool
    """
    if checkUpdate:
        checkForUpdate()

    show_proxy(session)

    banner()

    process_list = updateProcessList(session)
    if len(process_list) > 0:
        # Insert table header
        table = process_list.copy()
        table.insert(
            0, {"pid": "pid", "action": "task", "date": "date", "status": "status"}
        )
        # Get max length of strings in each category (date is always going to be 15)
        maxPid, maxAction, maxStatus = [
            max(i)
            for i in [
                [len(str(r["pid"])) for r in table],
                [len(str(r["action"])) for r in table],
                [len(str(r["status"])) for r in table],
            ]
        ]
        # Print header
        print(
            "|{:^{maxPid}}|{:^{maxAction}}|{:^15}|{:^{maxStatus}}|".format(
                table[0]["pid"],
                table[0]["action"],
                table[0]["date"],
                table[0]["status"],
                maxPid=maxPid,
                maxAction=maxAction,
                maxStatus=maxStatus,
            )
        )
        # Print process list
        [
            print(
                "|{:^{maxPid}}|{:^{maxAction}}|{:^15}|{:^{maxStatus}}|".format(
                    r["pid"],
                    r["action"],
                    datetime.datetime.fromtimestamp(r["date"]).strftime(
                        "%b %d %H:%M:%S"
                    ),
                    r["status"],
                    maxPid=maxPid,
                    maxAction=maxAction,
                    maxStatus=maxStatus,
                )
            )
            for r in process_list
        ]
        print("")

    menu_actions = {
        1: constructionList,
        2: sendResources,
        3: distributeResources,
        4: getStatus,
        5: activateShrine,
        6: loginDaily,
        701: alertAttacks,
        702: alertLowWine,
        801: buyResources,
        802: sellResources,
        901: donate,
        902: donationBot,
        10: vacationMode,
        11: activateMiracle,
        1201: trainArmy,
        1202: stationArmy,
        13: shipMovements,
        14: constructBuilding,
        15: update,
        16: webServer,
        17: autoPirate,
        18: investigate,
        1901: attackBarbarians,
        1902: autoBarbarians,
        2001: searchForIslandSpaces,
        2002: dumpWorld,
        2101: proxyConf,
        2102: updateTelegramData,
        2103: killTasks,
        2104: decaptchaConf,
        2105: logs,
        2106: testTelegramBot,
        2107: importExportCookie,
        2108: loadCustomModule,
        22: consolidateResources,
        23: modifyProduction
    }

    print("(0)  Exit")
    print("(1)  Construction list")
    print("(2)  Send resources")
    print("(3)  Distribute resources")
    print("(4)  Account status")
    print("(5)  Activate Shrine")
    print("(6)  Login daily")
    print("(7)  Alerts / Notifications")
    print("(8)  Marketplace")
    print("(9)  Donate")
    print("(10) Activate vacation mode")
    print("(11) Activate miracle")
    print("(12) Military actions")
    print("(13) See movements")
    print("(14) Construct building")
    print("(15) Update Ikabot")
    print("(16) Ikabot Web Server")
    print("(17) Auto-Pirate")
    print("(18) Investigate")
    print("(19) Attack / Grind barbarians")
    print("(20) Dump / Monitor world")
    print("(21) Options / Settings")
    print("(22) Consolidate resources")
    print("(23) Set Production of Saw mill / Luxury good")

    total_options = len(menu_actions) + 1
    selected = read(min=0, max=total_options, digit=True, empty=True)
    
    # refresh main menu on hitting enter
    if selected == '':
        return menu(session)

    if selected == 7:
        banner()
        print("(0) Back")
        print("(1) Alert attacks")
        print("(2) Alert wine running out")

        selected = read(min=0, max=2, digit=True)
        if selected == 0:
            menu(session)
            return
        if selected > 0:
            selected += 700

    if selected == 8:
        banner()
        print("(0) Back")
        print("(1) Buy resources")
        print("(2) Sell resources")

        selected = read(min=0, max=2, digit=True)
        if selected == 0:
            menu(session)
            return
        if selected > 0:
            selected += 800

    if selected == 9:
        banner()
        print("(0) Back")
        print("(1) Donate once")
        print("(2) Donate automatically")

        selected = read(min=0, max=2, digit=True)
        if selected == 0:
            menu(session)
            return
        if selected > 0:
            selected += 900

    if selected == 19:
        banner()
        print("(0) Back")
        print("(1) Simple Attack")
        print("(2) Auto Grind")
        selected = read(min=0, max=2, digit=True)
        if selected == 0:
            menu(session)
            return
        if selected > 0:
            selected += 1900

    if selected == 12:
        banner()
        print("(0) Back")
        print("(1) Train Army")
        print("(2) Send Troops/Ships")
        selected = read(min=0, max=2, digit=True)
        if selected == 0:
            menu(session)
            return
        if selected > 0:
            selected += 1200

    if selected == 20:
        print("(0) Back")
        print("(1) Monitor islands")
        print("(2) Dump & Search world")
        
        selected = read(min=0, max=2, digit=True)
        if selected == 0:
            menu(session)
            return
        if selected > 0:
            selected += 2000

    if selected == 21:
        banner()
        print("(0) Back")
        print("(1) Configure Proxy")
        if telegramDataIsValid(session):
            print("(2) Change the Telegram data")
        else:
            print("(2) Enter the Telegram data")
        print("(3) Kill tasks")
        print("(4) Configure captcha resolver")
        print("(5) Logs")
        print("(6) Message Telegram Bot")
        print("(7) Import / Export cookie")
        print("(8) Load custom ikabot module")

        selected = read(min=0, max=8, digit=True)
        if selected == 0:
            menu(session)
            return
        if selected > 0:
            selected += 2100

    if selected != 0:
        try:
            event = multiprocessing.Event()  # creates a new event
            config.has_params = len(config.predetermined_input) > 0
            process = multiprocessing.Process(
                target=menu_actions[selected],
                args=(session, event, sys.stdin.fileno(), config.predetermined_input),
                name=menu_actions[selected].__name__,
            )
            process.start()
            process_list.append(
                {
                    "pid": process.pid,
                    "action": menu_actions[selected].__name__,
                    "date": time.time(),
                    "status": "started",
                }
            )
            updateProcessList(session, programprocesslist=process_list)
            event.wait()  # waits for the process to fire the event that's been given to it. When it does  this process gets back control of the command line and asks user for more input
        except KeyboardInterrupt:
            pass
        menu(session, checkUpdate=False)
    else:
        if isWindows:
            # in unix, you can exit ikabot and close the terminal and the processes will continue to execute
            # in windows, you can exit ikabot but if you close the terminal, the processes will die
            print("Closing this console will kill the processes.")
            enter()
        clear()
        os._exit(
            0
        )  # kills the process which executes this statement, but it does not kill it's child processes


def init():
    home = "USERPROFILE" if isWindows else "HOME"
    os.chdir(os.getenv(home))
    if not os.path.isfile(ikaFile):
        open(ikaFile, "w")
        os.chmod(ikaFile, 0o600)


def start():
    init()
    config.has_params = len(sys.argv) > 1
    for arg in sys.argv:
        try:
            config.predetermined_input.append(int(arg))
        except ValueError:
            config.predetermined_input.append(arg)
    config.predetermined_input.pop(0)

    session = Session()
    try:
        menu(session)
    finally:
        clear()
        session.logout()


def main():
    manager = multiprocessing.Manager()
    predetermined_input = manager.list()
    config.predetermined_input = predetermined_input
    try:
        start()
    except KeyboardInterrupt:
        clear()


if __name__ == "__main__":
    # On Windows calling this function is necessary.
    if sys.platform.startswith("win"):
        multiprocessing.freeze_support()
    main()

#############################################################
# This is necessary to ensure that flask is frozen together #
# with other requirements when creating ikabot.exe          #
try: import flask                                           #
except: pass                                                #
#############################################################



#! /usr/bin/env python3
# -*- coding: utf-8 -*-

import ast
import gzip
import json
import os
import re
import sys
import threading
import time
import traceback
from pathlib import Path
import ikabot.config as config
from ikabot.config import *
from ikabot.helpers.botComm import sendToBot
from ikabot.helpers.getJson import getIsland
from ikabot.helpers.gui import banner, bcolors, enter
from ikabot.helpers.pedirInfo import read
from ikabot.helpers.process import set_child_mode
from ikabot.helpers.signals import setInfoSignal
from ikabot.helpers.varios import getDateTime, wait

LINE_UP = "\033[1A"
LINE_CLEAR = "\x1b[2K"
#              status, history, start_time
stop_updating = threading.Event()
lock = threading.Lock()
shared_data = ["", "", 0, stop_updating, lock]
home = "USERPROFILE" if isWindows else "HOME"
selected_islands = set()


def dumpWorld(session, event, stdin_fd, predetermined_input):
    """
    Parameters
    ----------
    session : ikabot.web.session.Session
    event : multiprocessing.Event
    stdin_fd: int
    predetermined_input : multiprocessing.managers.SyncManager.list
    """
    sys.stdin = os.fdopen(stdin_fd)
    config.predetermined_input = predetermined_input

    try:
        banner()
        if os.path.exists(os.getenv(home) + "/ikabot_world_dumps"):
            print("1) Create new dump")
            print("2) Load existing dump")
            choice = read(min=1, max=2, digit=True)
            if choice == 2:
                view_dump(session, event)
                event.set()
                return
        banner()
        print(
            "{}⚠️ BEWARE - THE RESULTING DUMP CONTAINS ACCOUNT IDENTIFYING INFORMATION ⚠️{}\n".format(
                bcolors.WARNING, bcolors.ENDC
            )
        )
        print(
            "This action will take a couple of hours to complete. Are you sure you want to initiate a data dump now? (Y|N)"
        )
        choice = read(values=["y", "Y", "n", "N"])
        if choice in ["n", "N"]:
            event.set()
            return
        print(
            "Type in the waiting time between each request in miliseconds (default = 1500): "
        )
        choice = read(min=0, max=10000, digit=True, default=1500)
        waiting_time = int(choice) / 1000
        print(
            "Do you want only shallow data about the islands? If yes you will not be able to search the dump by player names but the dump will be quick. (Y|N): "
        )
        choice = read(values=["y", "Y", "n", "N"])
        shallow = choice in ["y", "Y"]
        coords = None
        radius = None
        non_empty_islands = False
        if not shallow:
            print("Do you want to only dump a part of the map? (Y|N)")
            choice = read(values=["y", "Y", "n", "N"])
            if choice in ["y", "Y"]:
                print('Type in a center point (x,y): (type "skip" to skip this step)')
                input = read()
                if input.strip().lower() != "skip":
                    coords = input.replace("(", "").replace(")", "").split(",")
                    coords = (int(coords[0]), int(coords[1]))
                    print(
                        "Type in a max distance from the center point: (default = 15)"
                    )
                    radius = read(min=0, max=200, digit=True, default=15)
                print("Do you want to only dump islands with at least 1 town? (Y|N)")
                choice = read(values=["y", "Y", "n", "N"])
                non_empty_islands = choice in ["y", "Y"]

        thread = threading.Thread(target=update_terminal, args=(shared_data,), daemon=True)
        thread.start()
        set_child_mode(session)
        info = "\nDumped world data\n"
        setInfoSignal(session, info)

        dump_path = do_it(
            session, waiting_time, coords, radius, shallow, non_empty_islands
        )

        shared_data[3].set()
        shared_data[4].acquire()
        time.sleep(5)

        banner()
        print(
            "\n{}SUCCESS!{} World data has been dumped to {} in {}s \n".format(
                bcolors.GREEN,
                bcolors.ENDC,
                dump_path,
                str(round(time.time() - shared_data[2])),
            )
        )
        enter()
        event.set()
        return
    except Exception:
        shared_data[3].set()
        shared_data[4].acquire(timeout=10)
        event.set()
        msg = "Error in:\n{}\nCause:\n{}".format(info, traceback.format_exc())
        sendToBot(session, msg)
        return


def do_it(session, waiting_time, coords, radius, shallow, non_empty_islands):
    """
    Parameters
    ----------
    session : ikabot.web.session.Session
        Session object
    waiting_time : int
        Time to wait between each network request (to prevent getting rate limited)
    start_id : int
        Id of the island to start the dump from (0 starts from beginning)
    shallow : str
        String that determines if the data should be shallow (only map accessible data)

    Returns
    -------
    dump_path: str
    """

    shared_data[2] = time.time()
    partial = (
        True if (coords and radius and radius >= 0) or non_empty_islands else False
    )
    world = {
        "name": "s" + str(session.mundo) + "-" + str(session.servidor),
        "self_name": session.username,
        "dump_start_date": time.time(),
        "dump_end_date": 0,
        "islands": [],
        "shallow": shallow,
        "partial": partial,
    }
    shared_data.append(world)
    # scan 0 to 50 x and y
    shallow_islands = []
    update_status("Initiating first map sweep", 0, 0, True)
    update_status("Getting (0-50,0-50) islands", 25, 1.25)
    data = session.post(
        "action=WorldMap&function=getJSONArea&x_min=0&x_max=50&y_min=0&y_max=50"
    )
    for x, val in json.loads(data)["data"].items():
        for y, val2 in val.items():
            shallow_islands.append(
                {
                    "x": x,
                    "y": y,
                    "id": val2[0],
                    "name": val2[1],
                    "resource_type": val2[2],
                    "miracle_type": val2[3],
                    "wood_lvl": val2[6],
                    "players": val2[7],
                }
            )
    update_status("Getting (50-100,0-50) islands", 50, 2.5)
    time.sleep(0.5)
    data = session.post(
        "action=WorldMap&function=getJSONArea&x_min=50&x_max=100&y_min=0&y_max=50"
    )
    for x, val in json.loads(data)["data"].items():
        for y, val2 in val.items():
            shallow_islands.append(
                {
                    "x": x,
                    "y": y,
                    "id": val2[0],
                    "name": val2[1],
                    "resource_type": val2[2],
                    "miracle_type": val2[3],
                    "wood_lvl": val2[6],
                    "players": val2[7],
                }
            )
    update_status("Getting (0-50,50-100) islands", 75, 3.75)
    time.sleep(0.5)
    data = session.post(
        "action=WorldMap&function=getJSONArea&x_min=0&x_max=50&y_min=50&y_max=100"
    )
    for x, val in json.loads(data)["data"].items():
        for y, val2 in val.items():
            shallow_islands.append(
                {
                    "x": x,
                    "y": y,
                    "id": val2[0],
                    "name": val2[1],
                    "resource_type": val2[2],
                    "miracle_type": val2[3],
                    "wood_lvl": val2[6],
                    "players": val2[7],
                }
            )
    update_status("Getting (50-100,50-100) islands", 100, 5)
    time.sleep(0.5)
    data = session.post(
        "action=WorldMap&function=getJSONArea&x_min=50&x_max=100&y_min=50&y_max=100"
    )
    for x, val in json.loads(data)["data"].items():
        for y, val2 in val.items():
            shallow_islands.append(
                {
                    "x": x,
                    "y": y,
                    "id": val2[0],
                    "name": val2[1],
                    "resource_type": val2[2],
                    "miracle_type": val2[3],
                    "wood_lvl": val2[6],
                    "players": val2[7],
                }
            )

    # [
    # "58",         //id 0
    # "Phytios",    //name 1
    # "1",          //resource type 2
    # "2",          //type of miracle 3
    # "5",          // ?? 4
    # "4",          // ?? 5
    # "9",          // lumber level  6
    # "12",         // number of people 7
    # 0,            // piracy in range 8
    # "0",          // helios tower 9
    # "0",          // red 10
    # "0"           // blue 11
    # ]

    dump_path = (
        os.getenv(home)
        + "/ikabot_world_dumps/s"
        + str(session.mundo)
        + "-"
        + str(session.servidor)
        + "/"
    )
    dump_path = dump_path.replace("\\", "/")
    dump_name = getDateTime() + ".json.gz"

    if shallow:
        dump_name = dump_name.replace(".json.gz", "_shallow") + ".json.gz"
        update_status("Shallow dump is on. Dumping data...", 100, 100, True)
        world["islands"] = shallow_islands
        dump(world, dump_path, dump_name)
        return dump_path + dump_name

    all_island = set()
    total_settlements = 0
    for island in shallow_islands:
        island_id = int(island["id"])
        x = int(island["x"])
        y = int(island["y"])
        if non_empty_islands and not int(island["players"]):
            continue
        if (
            coords
            and radius
            and ((x - coords[0]) ** 2 + (y - coords[1]) ** 2) ** 0.5 > radius
        ):
            continue
        all_island.add((island_id, x, y))
        total_settlements += int(island["players"])

    update_status(
        "Got {} islands with {} towns in total".format(
            len(all_island), str(total_settlements)
        ),
        100,
        5,
        True,
    )
    update_status("Getting data for each island. This will take a while...", 0, 5, True)

    # scan each island

    all_island = sorted(all_island)
    dump_islands(shared_data, all_island, waiting_time, session)
    update_status("Got {} individual islands".format(len(all_island)), 100, 100, True)

    name_suffix = "_partial" if partial else ""
    dump_name = dump_name.replace(".json.gz", name_suffix) + ".json.gz"
    update_status("Dumping data to {}".format(dump_path + dump_name), 100, 100, True)
    dump(shared_data[5], dump_path, dump_name)
    return dump_path + dump_name


def dump(world, dump_path, dump_name):
    world["dump_end_date"] = time.time()
    p = Path(dump_path)
    p.mkdir(exist_ok=True, parents=True)
    with gzip.open(dump_path + dump_name, "wb") as file:
        json_string = json.dumps(world).encode("utf-8")
        file.write(json_string)


def dump_islands(shared_data, all_island, waiting_time, session):
    world_islands_number = len(all_island)
    for island in all_island:
        island_id = island[0]
        island_coords = (island[1], island[2])
        update_status(
            "Getting island id: {}, x: {}, y: {}".format(
                island_id, island_coords[0], island_coords[1]
            ),
            len(shared_data[5]["islands"]) / world_islands_number * 100,
            5 + (len(shared_data[5]["islands"]) / world_islands_number * 95),
        )
        html = ""
        try:
            html = session.get("view=island&islandId=" + str(island_id))
        except Exception:
            # try again
            html = session.get("view=island&islandId=" + str(island_id))
        island = getIsland(html)
        shared_data[4].acquire()
        shared_data[5]["islands"].append(island)
        shared_data[4].release()
        time.sleep(waiting_time)


def update_terminal(shared_data):
    while True:
        banner()
        print("\n")
        shared_data[4].acquire()
        print(shared_data[1])
        shared_data[4].release()
        chars = ["\\", "|", "/", "─"]
        for i in range(20):
            shared_data[4].acquire()
            print(" " * 120, end="\r")
            print(
                shared_data[0]
                + ",\tdt: "
                + str(round(time.time() - shared_data[2], 2))
                + "s\t"
                + chars[i % 4],
                end="\r",
            )
            shared_data[4].release()
            time.sleep(0.05)
        if stop_updating.is_set():
            return


def update_status(message, percent, percent_total, add_history=False):
    shared_data[4].acquire()
    shared_data[0] = (
        message
        + "\t"
        + str(round(percent, 1))
        + "%,\ttotal: "
        + str(round(percent_total, 1))
        + "%"
    )
    if add_history:
        shared_data[1] += shared_data[0] + "\n"
    shared_data[4].release()


def view_dump(session, event):


    files = [
        file.replace("\\", "/")
        for file in get_files(os.getenv(home) + "/ikabot_world_dumps")
        if ".json.gz" in file
    ]

    print("All dumps are stored in " + os.getenv(home) + "/ikabot_world_dumps\n")
    print("Choose a dump to view:")
    for i, file in enumerate(files):
        print(
            str(i)
            + ") "
            + file.split("/")[-2]
            + " "
            + file.split("/")[-1].replace(".json.gz", "").replace("_", " ")
        )
    choice = read(min=0, max=len(files) - 1, digit=True)
    print("Loading dump...")
    selected_dump = files[choice]
    with gzip.open(selected_dump, "rb") as file:
        selected_dump = json.load(file)

    while True:
        banner()
        print_map(selected_dump["islands"])
        print("0) Back")
        print("1) Search islands by island criteria")
        if not selected_dump["shallow"]:
            print("2) Search islands by player name")
            print("3) Search for nearest inactive players")
            print("4) Search for alliances")
            print("5) Search for a specific island by coordinates")

        choice = read(min=0, max=5, digit=True)
        if choice == 0:
            event.set()
            return
        elif choice == 1:
            print(
                "Search island by a certain criteria. The available properties of an island are:"
            )
            print(
                "resource_type : [1,2,3,4] // these are  Wine, Marble, Cristal, Sulfur"
            )
            print("miracle_type : [1,2,3,4,5,6,7,8] // hephaistos forge is number 5")
            print("wood_lvl : [1..] // this is the forest level on the island")
            print(
                "luxury_lvl : [1..] // this is the level of the luxury resource on the island (Only available in full dumps)"
            )
            print("players : [0..] // number of players on the island")
            print(
                "ex. If I wanted to find all islands with less than 10 players and forest level 30 with hephaistos I would type in:"
            )
            print("players < 10 and wood_lvl == 30 and miracle_type == 5\n")
            condition = read(msg="Enter the condition: ")

            try:
                filtered_islands = [
                    island
                    for island in filter(
                        lambda x: filter_on_condition(x, condition),
                        (
                            selected_dump["islands"]
                            if selected_dump["shallow"]
                            else convert_to_shallow(selected_dump["islands"])
                        ),
                    )
                ]
            except (SyntaxError, KeyError):
                print(
                    "Condition is bad, please use only the available island properties and use python standard conditional sytnax (and, or, <, >, ==, (, ), etc... )\nRemember luxury_lvl is only available in full dumps!"
                )
                enter()
                continue

            print("The satisfying islands are:")
            [print(island) for island in filtered_islands]
            enter()
        elif choice == 2:
            if selected_dump["shallow"]:
                print(
                    "You can not search by player name because this dump is shallow and doesn't contain data about players!"
                )
                enter()
                continue
            player_name = read(msg="Type in the player name: ")
            # search for players by name
            players = []
            for island in selected_dump["islands"]:
                for city in island["cities"]:
                    if city["type"] != "empty" and player_name in city["Name"]:
                        players.append((player_name, island["id"]))
            # return if none are found
            if not len(players):
                print("No players found!")
                enter()
                continue
            # select one
            print("Chose a player to add to selection: ")
            for i, player in enumerate(unique_tuples(players)):
                print(str(i) + ") " + player[0])
            choice = read(min=0, max=len(list(unique_tuples(players))) - 1, digit=True)
            # add his islands to selection
            for player in players:
                if player[0] == list(unique_tuples(players))[choice][0]:
                    selected_islands.add(int(player[1]))
        elif choice == 3:
            if selected_dump["shallow"]:
                print(
                    "You can not search by player name because this dump is shallow and doesn't contain data about players!"
                )
                enter()
                continue
            coords = (
                read(msg="Type in a center point (x,y): ")
                .replace("(", "")
                .replace(")", "")
                .split(",")
            )
            coords = (int(coords[0]), int(coords[1]))
            number_of_inactives = read(
                msg="How many inactives should be displayed? (min=1, default=25): ",
                min=1,
                digit=True,
                default=25,
            )
            # sort islands based on distance from center point
            islands_sorted = sorted(
                selected_dump["islands"],
                key=lambda island: (
                    (island["x"] - coords[0]) ** 2 + (island["y"] - coords[1]) ** 2
                )
                ** 0.5,
            )
            print("The nearest 25 inactive players are: ")
            # below follows the unholiest way to get the first n cities which are contained in an island object which is contained in a list of islands without duplicates in one line of code using python list comprehension
            seen = set()
            inactives = [
                city
                for island in islands_sorted
                for city in island["cities"]
                if city["type"] != "empty"
                and city["state"] == "inactive"
                and isinstance(
                    [(seen.add(city["Name"]),) if city["Name"] not in seen else None][
                        0
                    ],
                    tuple,
                )
            ][:number_of_inactives]
            for i, city in enumerate(inactives):
                print(str(i + 1) + ") " + city["Name"])
            enter()
        elif choice == 4:
            if selected_dump["shallow"]:
                print(
                    "You can not search by alliance tag because this dump is shallow and doesn't contain data about alliances!"
                )
                enter()
                continue
            while True:
                alliance = (read(msg="Enter the alliance TAG you want to search for (ONLY tags): ")).lower()
                user_islands = dict()
                alliance_found = False # Flag to check if any alliance is found
                for island in selected_dump['islands']:
                    for city in island["cities"]:
                        if "AllyTag" in city and city["AllyTag"].lower() == alliance:
                            user_name = city["Name"]
                            user_id = city["Id"]
                            if user_name not in user_islands:
                                user_islands[user_name] = {"id": user_id, "islands": set()}
                            user_islands[user_name]["islands"].add((island["x"], island["y"]))
                            alliance_found = True # Set FLag to True if alliance has been found
                if alliance_found:
                    for user_name, infe in user_islands.items():
                        print(f"Player:{bcolors.CYAN} {user_name}{bcolors.ENDC}, Alliance:{bcolors.CYAN} {alliance}{bcolors.ENDC}")
                        island_coordss = ", ".join([f"{x}:{y}" for x, y in infe['islands']])
                        print(f"- {bcolors.STONE}Islands{bcolors.ENDC}: {bcolors.YELLOW}{island_coordss}{bcolors.ENDC}")
                else:
                    print("No alliance with this tag was found.")
                # Ask if the user wants to search again
                search_again = input("Do you want to search for another alliance ? (yes/no): ").lower()

                if search_again != 'yes':
                    break  # Exit the loop if the user doesn't want to search again
        elif choice == 5:
            if selected_dump["shallow"]:
                print(
                    "You can not search by coordinates because this dump is shallow and doesn't contain full data about islands!"
                )
                enter()
                continue
            while True:
                coords = (read(msg="Enter the coordinates of the island you want to search for (xx:yy): "))
                try:
                    x, y = map(int, coords.split(':'))
                except ValueError:
                    print("Invalid input format. Please enter coordinates in the form xx:yy")
                    continue # Prompt again for the correct input
                island_found = False # Flag to check if the island has been found
                player_state_dict = {
                        "": "Active",
                        "inactive": "Inactive",
                        "vacation": "Vacation Mode"
                        }
                tradegood_dict = {
                        1: "Wine",
                        2: 'Marble',
                        3: "Crystal",
                        4: "Sulphur"
                        }
                wonder_dict = {
                        "1": "Hephaistos",
                        "2": "Hades",
                        "3": "Demeter",
                        "4": "Athene",
                        "5": "Hermes",
                        "6": "Ares",
                        "7": "Poseidon",
                        "8": "Colossus"
                        }
                for island in selected_dump['islands']:
                    if island["x"] == x and island["y"] == y:
                        player_data = {} # Dictionary to track players and their stats
                        # First count the number of cities for each player on the island
                        player_city_count = {}
                        for city in island['cities']:
                            if city['type'] == 'city':
                                player_id = city['Id']
                                if player_id not in player_city_count:
                                    player_city_count[player_id] = 0
                                player_city_count[player_id] += 1
                        # Now process each city and collect the needed data
                        for city in island['cities']:
                            if city['type'] == 'city':
                                player_name = city['Name']
                                player_id = city['Id']
                                player_state = player_state_dict.get(city['state'])
                                player_alliance = city.get('AllyTag', 'No Alliance')
                                if player_id not in player_data: # Add only if player ID is not already in the dictionary
                                    player_score_raw = island['avatarScores'].get(player_id, {}).get('building_score_main')
                                    player_score_int = int(player_score_raw.replace(',', ''))
                                    player_score = "{:,}".format(player_score_int // 100)
                                    city_count = player_city_count[player_id]  # Get the number of cities for this player
                                    player_data[player_id] = f"{bcolors.CYAN}{player_name}{bcolors.ENDC} - Score: {bcolors.YELLOW}{player_score}{bcolors.ENDC}, Cities: {bcolors.YELLOW}{city_count}{bcolors.ENDC}, State: {bcolors.CYAN}{player_state}{bcolors.ENDC}, Alliance: {bcolors.CYAN}{player_alliance}{bcolors.ENDC}"
                        isl_name = island['name']
                        cities_of_type_city = [city for city in island['cities'] if city['type'] == 'city']
                        sum_cities = len(cities_of_type_city) if cities_of_type_city else 0
                        wood_lvl = island['resourceLevel']
                        luxury_lvl = island['tradegoodLevel']
                        isl_type = tradegood_dict.get(island['tradegood']) 
                        isl_id = island['id']
                        wonder = wonder_dict[island['wonder']]
                        wonder_lvl = island['wonderLevel']
                        island_found = True # Set flag to True if island has been found
                        break  # Exit the loop once the island is found
                if island_found:
                    print(
                        f"Island Name: {bcolors.CYAN}{isl_name}{bcolors.ENDC}, Number of cities: {bcolors.YELLOW}{sum_cities}{bcolors.ENDC}, Wonder: {bcolors.CYAN}{wonder}{bcolors.ENDC}, Wonder Level: {bcolors.YELLOW}{wonder_lvl}{bcolors.ENDC}")
                    print(
                        f"Forrest Level: {bcolors.YELLOW}{wood_lvl}{bcolors.ENDC}, Luxury Mine Level: {bcolors.YELLOW}{luxury_lvl}{bcolors.ENDC}, Island ID: {bcolors.YELLOW}{isl_id}{bcolors.ENDC}, Island Type: {bcolors.CYAN}{isl_type}{bcolors.ENDC}")
                    if not player_data:
                        print(f"No players in this island")
                    else:
                        player_data_str = '\n'.join(player_data.values())
                        print(f"Players:\n{player_data_str}")
                else:
                    print(f"No such island was found.")
                # Ask if the user wants to search again
                search_again = input("Do you want to search for another island ? (yes/no): ").lower()
                if search_again != 'yes':
                    break  # Exit the loop if the user doesn't want to search again
                    



def print_map(islands):
    """Prints out a 100x100 matrix with all world islands on it. Selected islands are colored red.
    Parameters
    ----------
    islands : [object]
        List of island objects to be displayed
    """

    map = [
        [bcolors.DARK_BLUE + "██" + bcolors.ENDC for j in range(100)]
        for i in range(100)
    ]  # 100x100 matrix of dark blue ██
    selected_island_coords = []

    for island in islands:
        if int(island["id"]) in selected_islands:
            map[int(island["x"]) - 1][int(island["y"]) - 1] = (
                bcolors.DARK_RED + "◉ " + bcolors.ENDC
            )
            selected_island_coords.append((int(island["x"]), int(island["y"])))
        else:
            map[int(island["x"]) - 1][int(island["y"]) - 1] = (
                bcolors.DARK_GREEN + "◉ " + bcolors.ENDC
            )

    for row in reversed(map):
        print("".join(row))

    print(
        bcolors.DARK_BLUE
        + "██"
        + bcolors.ENDC
        + " - Water, "
        + bcolors.DARK_GREEN
        + "◉"
        + bcolors.ENDC
        + " - Island, "
        + bcolors.DARK_RED
        + "◉"
        + bcolors.ENDC
        + " - Selected\n"
    )

    print("Selected islands: " + str(selected_island_coords))


def filter_on_condition(island, condition):
    """Returns true if island satisfies condition
    Parameters
    ----------
    island : object
        Island to be tested on condition
    condition : str
        String that represents a valid python condition to be applied to filter the list of islands

    Returns
    -------
    is_satisfied : bool
        Bool indicating whether or not the island object satisfies the condition
    """

    condition = ast.parse(condition)
    for node in ast.walk(condition):
        if isinstance(node, ast.Compare):
            left = node.left.id
            right = (
                node.comparators[0].n
                if isinstance(node.comparators[0], ast.Num)
                else node.comparators[0].id
            )
            op = node.ops[0]
            if op.__class__ == ast.Lt:
                if not int(island[left]) < int(right):
                    return False
            elif op.__class__ == ast.Gt:
                if not int(island[left]) > int(right):
                    return False
            elif op.__class__ == ast.Eq:
                if not int(island[left]) == int(right):
                    return False
    return True


def convert_to_shallow(islands):
    """Converts a list of islands from a deep dump into a shallow dump version of that list
    Parameters
    ----------
    islands : [object]
        List of island objects to be converted

    Returns
    -------
    islands : [object]
        List of objects that represent the stripped-down version of an island
    """
    return [
        {
            "x": str(island["x"]),
            "y": str(island["y"]),
            "id": island["id"],
            "name": island["name"],
            "resource_type": island["tradegood"],
            "miracle_type": island["wonder"],
            "wood_lvl": island["resourceLevel"],
            "luxury_lvl": island["tradegoodLevel"],
            "players": len(
                [city for city in island["cities"] if city["type"] != "empty"]
            ),
        }
        for island in islands
    ]


def unique_tuples(tuples):
    """Iterates over tuples with a unique first element"""
    seen = {}
    for t in tuples:
        if t[0] not in seen:
            seen[t[0]] = True
            yield t


def get_files(path):
    """
    Returns all full paths to every file in a directory and all it's subdirectories
    Parameters
    ----------
    path : str
        Path to directory
    Returns
    -------
    files : list
    """
    files = []
    for dirpath, dirnames, filenames in os.walk(path):
        for filename in filenames:
            files.append(os.path.join(dirpath, filename))
    return files



#! /usr/bin/env python3
# -*- coding: utf-8 -*-

import json
import math
import time
import traceback
from decimal import *

from ikabot.config import *
from ikabot.helpers.botComm import *
from ikabot.helpers.gui import *
from ikabot.helpers.market import *
from ikabot.helpers.naval import getTotalShips
from ikabot.helpers.pedirInfo import read
from ikabot.helpers.planRoutes import waitForArrival
from ikabot.helpers.process import set_child_mode
from ikabot.helpers.signals import setInfoSignal
from ikabot.helpers.varios import addThousandSeparator, wait, getDateTime
from ikabot.helpers.pedirInfo import getShipCapacity


def chooseCommercialCity(commercial_cities):
    """
    Parameters
    ----------
    commercial_cities : list[dict]

    Returns
    -------
    commercial_city : dict
    """
    print("In which city do you want to sell resources?\n")
    for i, city in enumerate(commercial_cities):
        print("({:d}) {}".format(i + 1, city["name"]))
    ind = read(min=1, max=len(commercial_cities))
    return commercial_cities[ind - 1]


def getMarketInfo(session, city):
    """
    Parameters
    ----------
    session : ikabot.web.session.Session
    city : dict

    Returns
    -------
    response : dict
    """
    params = {
        "view": "branchOfficeOwnOffers",
        "activeTab": "tab_branchOfficeOwnOffers",
        "cityId": city["id"],
        "position": city["pos"],
        "backgroundView": "city",
        "currentCityId": city["id"],
        "templateView": "branchOfficeOwnOffers",
        "currentTab": "tab_branchOfficeOwnOffers",
        "actionRequest": actionRequest,
        "ajax": "1",
    }
    resp = session.post(params=params, noIndex=True)
    return json.loads(resp, strict=False)[1][1][1]


def getOffers(session, my_market_city, resource_type):
    """
    Parameters
    ----------
    session : ikabot.web.session.Session
    my_market_city : dict
    resource_type : int

    Returns
    -------
    offers : list
    """
    if resource_type == 0:
        resource_type = "resource"
    else:
        resource_type = str(resource_type)
    data = {
        "cityId": my_market_city["id"],
        "position": my_market_city["pos"],
        "view": "branchOffice",
        "activeTab": "bargain",
        "type": "333",
        "searchResource": resource_type,
        "range": my_market_city["rango"],
        "backgroundView": "city",
        "currentCityId": my_market_city["id"],
        "templateView": "branchOffice",
        "currentTab": "bargain",
        "actionRequest": actionRequest,
        "ajax": "1",
    }
    resp = session.post(params=data)
    html = json.loads(resp, strict=False)[1][1][1]

    offers_found = re.findall(
        r'<td class="short_text80">(.*?)<br/>\((.*?)\)[\s\S]*?<div class="tooltip">([\d,]+)</div>[\s\S]*?<td style="white-space:nowrap;">(\d+)[\s\S]*?<td>(\d+)</td>[\s\S]*?href="\?view=takeOffer&destinationCityId=(\d+)',
        html
    )
    # Process the found offers, removing commas and converting to integers
    processed_offers = []
    for city_name, user_name, amount, price, dist, dest_id in offers_found:
        clean_amount = int(amount.replace(',', ''))
        processed_offers.append((city_name.strip(), user_name.strip(), clean_amount, int(price), int(dist), int(dest_id)))
    
    return processed_offers


def sellToOffers(session, city_to_buy_from, resource_type, event):
    """
    Parameters
    ----------
    session : ikabot.web.session.Session
    city_to_buy_from : dict
    resource_type : int
    event : multiprocessing.Event
    """
    banner()

    offers = getOffers(session, city_to_buy_from, resource_type)

    if len(offers) == 0:
        print("No offers available.")
        enter()
        event.set()
        return

    print("Which offers do you want to sell to?\n")

    chosen_offers = []
    total_amount = 0
    profit = 0
    for offer in offers:
        cityname, username, amount, price, dist, destination_city_id = offer
        cityname = cityname.strip()
        amount = int(amount)
        price = int(price)
        msg = "{} ({}): {} at {:d} each ({} in total) [Y/n]".format(
            cityname,
            username,
            addThousandSeparator(amount),
            price,
            addThousandSeparator(price * amount),
        )
        rta = read(msg=msg, values=["y", "Y", "n", "N", ""])
        if rta.lower() == "n":
            continue
        chosen_offers.append(offer)
        total_amount += amount
        profit += amount * price

    if len(chosen_offers) == 0:
        event.set()
        return

    available = city_to_buy_from["availableResources"][resource_type]
    amount_to_sell = min(available, total_amount)

    banner()
    print(
        "\nHow much do you want to sell? [max = {}]".format(
            addThousandSeparator(amount_to_sell)
        )
    )
    amount_to_sell = read(min=0, max=amount_to_sell)
    if amount_to_sell == 0:
        event.set()
        return

    left_to_sell = amount_to_sell
    profit = 0
    for offer in chosen_offers:
        cityname, username, amount, price, dist, destination_city_id = offer
        cityname = cityname.strip()
        amount = int(amount)
        price = int(price)
        sell = min(amount, left_to_sell)
        left_to_sell -= sell
        profit += sell * price
    print(
        "\nSell {} of {} for a total of {}? [Y/n]".format(
            addThousandSeparator(amount_to_sell),
            materials_names[resource_type],
            addThousandSeparator(profit),
        )
    )
    rta = read(values=["y", "Y", "n", "N", ""])
    if rta.lower() == "n":
        event.set()
        return

    set_child_mode(session)
    event.set()

    info = "\nI sell {} of {} in {}\n".format(
        addThousandSeparator(amount_to_sell),
        materials_names[resource_type],
        city_to_buy_from["name"],
    )
    setInfoSignal(session, info)
    try:
        do_it1(session, amount_to_sell, chosen_offers, resource_type, city_to_buy_from)
    except Exception as e:
        msg = "Error in:\n{}\nCause:\n{}".format(info, traceback.format_exc())
        sendToBot(session, msg)
    finally:
        session.logout()


def createOffer(session, my_offering_market_city, resource_type, event):
    """
    Parameters
    ----------
    session : ikabot.web.session.Session
    my_offering_market_city : dict
    resource_type : int
    event : multiprocessing.Event
    """
    banner()

    html = getMarketInfo(session, my_offering_market_city)
    sell_market_capacity = storageCapacityOfMarket(html)
    total_available_amount_of_resource = my_offering_market_city["availableResources"][
        resource_type
    ]

    print(
        "How much do you want to sell? [max = {}]".format(
            addThousandSeparator(total_available_amount_of_resource)
        )
    )
    amount_to_sell = read(min=0, max=total_available_amount_of_resource)
    if amount_to_sell == 0:
        event.set()
        return

    price_max, price_min = re.findall(r"\'upper\': (\d+),\s*\'lower\': (\d+)", html)[
        resource_type
    ]
    price_max = int(price_max)
    price_min = int(price_min)
    print("\nAt what price? [min = {:d}, max = {:d}]".format(price_min, price_max))
    price = read(min=price_min, max=price_max)

    print(
        "\nI will sell {} of {} at {}: {}".format(
            addThousandSeparator(amount_to_sell),
            materials_names[resource_type],
            addThousandSeparator(price),
            addThousandSeparator(price * amount_to_sell),
        )
    )
    print("\nProceed? [Y/n]")
    rta = read(values=["y", "Y", "n", "N", ""])
    if rta.lower() == "n":
        event.set()
        return

    set_child_mode(session)
    event.set()

    info = "\nI sell {} of {} in {}\n".format(
        addThousandSeparator(amount_to_sell),
        materials_names[resource_type],
        my_offering_market_city["name"],
    )
    setInfoSignal(session, info)
    try:
        do_it2(
            session,
            amount_to_sell,
            price,
            resource_type,
            sell_market_capacity,
            my_offering_market_city,
        )
    except Exception as e:
        msg = "Error in:\n{}\nCause:\n{}".format(info, traceback.format_exc())
        sendToBot(session, msg)
    finally:
        session.logout()


def sellResources(session, event, stdin_fd, predetermined_input):
    """
    Parameters
    ----------
    session : ikabot.web.session.Session
    event : multiprocessing.Event
    stdin_fd: int
    predetermined_input : multiprocessing.managers.SyncManager.list
    """
    sys.stdin = os.fdopen(stdin_fd)
    config.predetermined_input = predetermined_input
    try:
        banner()

        commercial_cities = getCommercialCities(session)
        if len(commercial_cities) == 0:
            print("There is no store built")
            enter()
            event.set()
            return

        if len(commercial_cities) == 1:
            city = commercial_cities[0]
        else:
            city = chooseCommercialCity(commercial_cities)
            banner()

        print("What resource do you want to sell?")
        for index, material_name in enumerate(materials_names):
            print("({:d}) {}".format(index + 1, material_name))
        selected_material = read(min=1, max=len(materials_names))
        resource = selected_material - 1
        banner()

        print(
            
            "Do you want to sell to existing offers (1) or do you want to make your own offer (2)?"
            
        )
        selected = read(min=1, max=2)
        [sellToOffers, createOffer][selected - 1](session, city, resource, event)
    except KeyboardInterrupt:
        event.set()
        return


def do_it1(session, left_to_sell, offers, resource_type, city_to_buy_from):
    """
    Parameters
    ----------
    session : ikabot.web.session.Session
    left_to_sell : int
    offers : list[dict]
    resource_type : int
    city_to_buy_from : dict
    """
    ship_capacity, freighter_capacity = getShipCapacity(session)
    for offer in offers:
        cityname, username, amount, precio, dist, destination_city_id = offer
        cityname = cityname.strip()
        amount_to_buy = int(amount)
        while True:
            amount_to_sell = min(amount_to_buy, left_to_sell)
            ships_available = waitForArrival(session)
            ships_needed = math.ceil((Decimal(amount_to_sell) / Decimal(ship_capacity)))
            ships_used = min(ships_available, ships_needed)
            if ships_needed > ships_used:
                amount_to_sell = ships_used * ship_capacity
            left_to_sell -= amount_to_sell
            amount_to_buy -= amount_to_sell

            data = {
                "action": "transportOperations",
                "function": "sellGoodsAtAnotherBranchOffice",
                "cityId": city_to_buy_from["id"],
                "destinationCityId": destination_city_id,
                "oldView": "branchOffice",
                "position": city_to_buy_from["pos"],
                "avatar2Name": username,
                "city2Name": cityname,
                "type": "333",
                "activeTab": "bargain",
                "transportDisplayPrice": "0",
                "premiumTransporter": "0",
                "normalTransportersMax": ships_available,
                "capacity": "5",
                "max_capacity": "5",
                "jetPropulsion": "0",
                "transporters": str(ships_used),
                "backgroundView": "city",
                "currentCityId": city_to_buy_from["id"],
                "templateView": "takeOffer",
                "currentTab": "bargain",
                "actionRequest": actionRequest,
                "ajax": "1",
            }
            if resource_type == 0:
                data["cargo_resource"] = amount_to_sell
                data["resourcePrice"] = precio
            else:
                data["tradegood{:d}Price".format(resource_type)] = precio
                data["cargo_tradegood{:d}".format(resource_type)] = amount_to_sell

            session.get(city_url + city_to_buy_from["id"], noIndex=True)
            session.post(params=data)
            
            resource_name = materials_names[resource_type]
            session.setStatus(
                f"Sold {addThousandSeparator(amount_to_sell)} {resource_name} for {precio} gold to {cityname} ({username}) <--- {city_to_buy_from['name']} | {getDateTime()}"
            )

            if left_to_sell == 0:
                return
            if amount_to_buy == 0:
                break


def do_it2(session, amount_to_sell, price, resource_type, sell_market_capacity, city):
    """
    Parameters
    ----------
    session : ikabot.web.session.Session
    amount_to_sell : int
    price : int
    resource_type : int
    sell_market_capacity : int
    city : dict
    """
    initial_amount_to_sell = amount_to_sell
    html = getMarketInfo(session, city)
    previous_on_sell = onSellInMarket(html)[resource_type]
    while True:
        html = getMarketInfo(session, city)
        currently_on_sell = onSellInMarket(html)[resource_type]
        # if there is space in the store
        if currently_on_sell < storageCapacityOfMarket(html):
            # add our new offer to the free space
            free_space = sell_market_capacity - currently_on_sell
            offer = min(amount_to_sell, free_space)
            amount_to_sell -= offer
            new_offer = currently_on_sell + offer

            payloadPost = {
                "cityId": city["id"],
                "position": city["pos"],
                "action": "CityScreen",
                "function": "updateOffers",
                "resourceTradeType": "444",
                "resource": "0",
                "resourcePrice": "10",
                "tradegood1TradeType": "444",
                "tradegood1": "0",
                "tradegood1Price": "11",
                "tradegood2TradeType": "444",
                "tradegood2": "0",
                "tradegood2Price": "12",
                "tradegood3TradeType": "444",
                "tradegood3": "0",
                "tradegood3Price": "17",
                "tradegood4TradeType": "444",
                "tradegood4": "0",
                "tradegood4Price": "5",
                "backgroundView": "city",
                "currentCityId": city["id"],
                "templateView": "branchOfficeOwnOffers",
                "currentTab": "tab_branchOfficeOwnOffers",
                "actionRequest": actionRequest,
                "ajax": "1",
            }
            if resource_type == 0:
                payloadPost["resource"] = new_offer
                payloadPost["resourcePrice"] = price
            else:
                payloadPost["tradegood{:d}".format(resource_type)] = new_offer
                payloadPost["tradegood{:d}Price".format(resource_type)] = price
            session.post(params=payloadPost)

            # if we don't have any more to add to the offer, leave the loop
            if amount_to_sell == 0:
                break

        # sleep for 2 hours
        wait(60 * 60 * 2)

    # wait until the last of our offer is actualy bought, and let the user know
    while True:
        html = getMarketInfo(session, city)
        currently_on_sell = onSellInMarket(html)[resource_type]
        if currently_on_sell <= previous_on_sell:
            msg = "{} of {} was sold at {:d}".format(
                addThousandSeparator(initial_amount_to_sell),
                materials_names[resource_type],
                price,
            )
            sendToBot(session, msg)
            return

        # sleep for 2 hours
        wait(60 * 60 * 2)



#! /usr/bin/env python3
# -*- coding: utf-8 -*-

from ikabot.helpers.logging import getLogger
logger = getLogger(__name__)
import socket
import struct

from ikabot.config import *
from ikabot.helpers.process import run


def getDNSTXTRecordWithSocket(domain, DNS_server="8.8.8.8"):
    """Returns the TXT record from the DNS server for the given domain
    Parameters
    ----------
    domain : str
        Domain name
    DNS_server : str
        DNS server address, default is '8.8.8.8'
    Returns
    -------
    str
        TXT record
    """

    # DNS Query
    def build_query(domain):
        # Header Section
        ID = struct.pack(">H", 0x1234)  # Identifier: transaction ID
        FLAGS = struct.pack(">H", 0x0100)  # Standard query with recursion
        QDCOUNT = struct.pack(">H", 0x0001)  # One question
        ANCOUNT = struct.pack(">H", 0x0000)  # No answers
        NSCOUNT = struct.pack(">H", 0x0000)  # No authority records
        ARCOUNT = struct.pack(">H", 0x0000)  # No additional records
        header = ID + FLAGS + QDCOUNT + ANCOUNT + NSCOUNT + ARCOUNT

        # Question Section
        question = b""
        for part in domain.split("."):
            question += struct.pack("B", len(part)) + part.encode("utf-8")
        question += struct.pack("B", 0)  # End of string
        QTYPE = struct.pack(">H", 0x0010)  # TXT record
        QCLASS = struct.pack(">H", 0x0001)  # IN class
        question += QTYPE + QCLASS

        return header + question

    # Send DNS Query
    def send_query(query, server=DNS_server, port=53):
        with socket.socket(socket.AF_INET, socket.SOCK_DGRAM) as sock:
            sock.settimeout(5)
            sock.sendto(query, (server, port))
            data, _ = sock.recvfrom(512)  # 512 bytes is the max size of DNS datagram
            return data

    # Parse DNS Response
    def parse_response(response):
        # Skip the header
        header_size = 12
        offset = header_size

        # Read the question section
        while True:
            length = response[offset]
            if length == 0:
                break
            offset += length + 1
        offset += 5  # Skip the zero byte and QTYPE + QCLASS

        # Read the answer section
        while offset < len(response):
            # Read the name
            if response[offset] == 0xC0:
                offset += 2  # Pointer to a name
            else:
                # Name in the form of a sequence of labels
                while True:
                    length = response[offset]
                    if length == 0:
                        break
                    offset += length + 1
                offset += 1  # End of the name

            type = struct.unpack(">H", response[offset : offset + 2])[0]
            offset += 10  # Type (2 bytes) + Class (2 bytes) + TTL (4 bytes) + Data length (2 bytes)

            if type == 16:  # TXT record
                txt_length = struct.unpack(">H", response[offset - 2 : offset])[0]
                txt_data = response[offset : offset + txt_length]
                # TXT records can be split into multiple strings
                txt_strings = []
                while txt_data:
                    string_length = txt_data[0]
                    txt_strings.append(txt_data[1 : string_length + 1].decode("utf-8"))
                    txt_data = txt_data[string_length + 1 :]
                return " ".join(txt_strings)
            else:
                # Skip this record and move to the next
                data_length = struct.unpack(">H", response[offset - 2 : offset])[0]
                offset += data_length

        raise ValueError("No TXT record found")

    query = build_query(domain)
    response = send_query(query)
    return "http://" + parse_response(response)


def getDNSTXTRecordWithNSlookup(domain, DNS_server="8.8.8.8"):
    """Returns the TXT record from the DNS server for the given domain using the nslookup tool
    Parameters
    ----------
    domain : str
        Domain name
    DNS_server : str
        DNS server address, default is '8.8.8.8'
    Returns
    -------
    str
        TXT record
    """
    text = run(f"nslookup -q=txt {domain} {DNS_server}")
    parts = text.split('"')
    if len(parts) < 2:
        # the DNS output is not well formed
        raise Exception(
            f'The command "nslookup -q=txt {domain} {DNS_server}" returned bad data: {text}'
        )
    return "http://" + parts[1]


def getAddressWithSocket(domain):
    """Makes multiple attempts to obtain the ikabot public API server address with the socket library
    Returns
    -------
    str
        server address
    """
    try:
        return getDNSTXTRecordWithSocket(domain, "ns2.afraid.org")
    except Exception as e:
        logger.warning("Failed to obtain public API address from ns2.afraid.org, trying with 8.8.8.8: ", exc_info=True)
    try:
        return getDNSTXTRecordWithSocket(domain, "8.8.8.8")
    except Exception as e:
        logger.warning("Failed to obtain public API address from 8.8.8.8, trying with 1.1.1.1: ", exc_info=True)
    try:
        return getDNSTXTRecordWithSocket(domain, "1.1.1.1")
    except Exception as e:
        logger.warning("Failed to obtain public API address from 1.1.1.1: ", exc_info=True)
        raise e


def getAddressWithNSlookup(domain):
    """Makes multiple attempts to obtain the ikabot public API server address with the nslookup tool if it's installed
    Returns
    -------
    str
        server address
    """
    try:
        return getDNSTXTRecordWithNSlookup(domain, "ns2.afraid.org")
    except Exception as e:
        logger.warning("Failed to obtain public API address from ns2.afraid.org: ", exc_info=True)
    try:
        return getDNSTXTRecordWithNSlookup(domain, "")
    except Exception as e:
        logger.warning("Failed to obtain public API address from nslookup with system default DNS server: ", exc_info=True)
    try:
        return getDNSTXTRecordWithNSlookup(domain, "8.8.8.8")
    except Exception as e:
        logger.warning("Failed to obtain public API address from nslookup with 8.8.8.8: ", exc_info=True)
    try:
        return getDNSTXTRecordWithNSlookup(domain, "1.1.1.1")
    except Exception as e:
        logger.warning("Failed to obtain public API address from nslookup with 1.1.1.1: ", exc_info=True)
        raise e


def getAddress(domain="ikagod.twilightparadox.com"):
    """Makes multiple attempts to obtain the ikabot public API server address
    Parameters
    ----------
    domain : str
        Domain name
    Returns
    -------
    str
        server address
    """
    custom_address = os.getenv("CUSTOM_API_ADDRESS")
    if custom_address:
        return custom_address
    try:
        address = getAddressWithSocket(domain)
        assert "." in address or ":" in address.replace("http://", ""), (
            "Bad server address: " + address
        )
        return address.replace("/ikagod/ikabot", "")
    except Exception as e:
        logger.warning("Failed to obtain public API address from socket, falling back to nslookup: ", exc_info=True)
    try:
        address = getAddressWithNSlookup(domain)
        assert "." in address or ":" in address.replace("http://", ""), (
            "Bad server address: " + address
        )  # address is either hostname, IPv4 or IPv6
        return address.replace("/ikagod/ikabot", "")
    except Exception as e:
        logger.error("Failed to obtain public API address from both socket and nslookup: ", exc_info=True)
        raise e



#! /usr/bin/env python3
# -*- coding: utf-8 -*-

import sys

import requests

import ikabot.config as config
from ikabot.helpers.gui import *
from ikabot.helpers.pedirInfo import read


def show_proxy(session):
    session_data = session.getSessionData()
    msg = "using proxy:"
    if "proxy" in session_data and session_data["proxy"]["set"] is True:
        curr_proxy = session_data["proxy"]["conf"]["https"]
        if test_proxy(session, session_data["proxy"]["conf"]) is False:
            session_data["proxy"]["set"] = False
            session.setSessionData(session_data)
            sys.exit(
                "the {} proxy does not work, it has been removed".format(curr_proxy)
            )
        if msg not in config.update_msg:
            # add proxy message
            config.update_msg += "{} {}\n".format(msg, curr_proxy)
        else:
            # delete old proxy message
            config.update_msg = config.update_msg.replace(
                "\n".join(config.update_msg.split("\n")[-2:]), ""
            )
            # add new proxy message
            config.update_msg += "{} {}\n".format(msg, curr_proxy)
    elif msg in config.update_msg:
        # delete old proxy message
        config.update_msg = config.update_msg.replace(
            "\n".join(config.update_msg.split("\n")[-2:]), ""
        )


def test_proxy(session, proxy_dict):
    try:
        requests.get(
            session.urlBase,
            proxies=proxy_dict,
            verify=config.do_ssl_verify,
        )
    except Exception as e:
        print('Proxy test failure. Error: ' + str(e))
        return False
    return True


def read_proxy(session):
    print(
        
        "Enter the proxy: protocol://username:password@address:port\n(examples: socks5://127.0.0.1:9050, https://45.117.163.22:8080):"
        
    )
    proxy_str = read(msg="proxy: ")
    proxy_dict = {"http": proxy_str, "https": proxy_str}
    if test_proxy(session, proxy_dict) is False:
        print("The proxy does not work.")
        enter()
        return None
    print("The proxy works and it will be used for all future requests sent by new ikabot processes.")
    enter()
    return proxy_dict


def proxyConf(session, event, stdin_fd, predetermined_input):
    """
    Parameters
    ----------
    session : ikabot.web.session.Session
    event : multiprocessing.Event
    stdin_fd: int
    predetermined_input : multiprocessing.managers.SyncManager.list
    """
    sys.stdin = os.fdopen(stdin_fd)
    config.predetermined_input = predetermined_input
    try:
        banner()
        print(
            "Warning: The proxy does not apply to the requests sent to the lobby!\n"
        )

        session_data = session.getSessionData()
        if "proxy" not in session_data or session_data["proxy"]["set"] is False:
            print("Right now, there is no proxy configured.")
            proxy_dict = read_proxy(session)
            if proxy_dict is None:
                event.set()
                return
            session_data["proxy"] = {}
            session_data["proxy"]["conf"] = proxy_dict
            session_data["proxy"]["set"] = True
        else:
            curr_proxy = session_data["proxy"]["conf"]["https"]
            print("Current proxy: {}".format(curr_proxy))
            print("What do you want to do?")
            print("0) Exit")
            print("1) Set a new proxy")
            print("2) Remove the current proxy")
            rta = read(min=0, max=2)

            if rta == 0:
                event.set()
                return
            if rta == 1:
                proxy_dict = read_proxy(session)
                if proxy_dict is None:
                    event.set()
                    return
                session_data["proxy"]["conf"] = proxy_dict
                session_data["proxy"]["set"] = True
            if rta == 2:
                session_data["proxy"]["set"] = False
                print("The proxy has been removed.")
                enter()

        session.setSessionData(session_data)
        event.set()
    except KeyboardInterrupt:
        event.set()
        return


#! /usr/bin/env python3
# -*- coding: utf-8 -*-

import json
import math
import re
import traceback
from decimal import *

from ikabot.config import *
from ikabot.helpers.botComm import *
from ikabot.helpers.getJson import getCity
from ikabot.helpers.gui import banner, enter
from ikabot.helpers.market import *
from ikabot.helpers.naval import getTotalShips
from ikabot.helpers.pedirInfo import getIdsOfCities, read
from ikabot.helpers.planRoutes import waitForArrival
from ikabot.helpers.process import set_child_mode
from ikabot.helpers.resources import *
from ikabot.helpers.signals import setInfoSignal
from ikabot.helpers.varios import addThousandSeparator, getDateTime
from ikabot.helpers.pedirInfo import getShipCapacity



def chooseResource(session, city):
    """
    Parameters
    ----------
    session : ikabot.web.session.Session
    city : dict
    """
    print("Which resource do you want to buy?")
    for index, material_name in enumerate(materials_names):
        print("({:d}) {}".format(index + 1, material_name))
    choise = read(min=1, max=5)
    resource = choise - 1
    if resource == 0:
        resource = "resource"
    data = {
        "cityId": city["id"],
        "position": city["pos"],
        "view": "branchOffice",
        "activeTab": "bargain",
        "type": 444,
        "searchResource": resource,
        "range": city["rango"],
        "backgroundView": "city",
        "currentCityId": city["id"],
        "templateView": "branchOffice",
        "currentTab": "bargain",
        "actionRequest": actionRequest,
        "ajax": 1,
    }
    # this will set the chosen resource in the store
    session.post(params=data)
    resource = choise - 1
    # return the chosen resource
    return resource


def getOffers(session, city):
    """
    Parameters
    ----------
    session : ikabot.web.session.Session
    city : dict
    Returns
    -------
    offers : list[dict]
    """
    html = getMarketHtml(session, city)
    hits = re.findall(
        r'short_text80">(.*?) <br/>\((.*?)\)\s*</td>\s*<td>(\d+)</td>\s*<td>([\d,\s]+)(?:<div class="tooltip">.*?)?</td>\s*<td><img src="([^"]+\.png)"[\s\S]*?white-space:nowrap;">(\d+)\s*[\s\S]*?href="\?view=takeOffer&destinationCityId=(\d+)&oldView=branchOffice&activeTab=bargain&cityId=(\d+)&position=(\d+)&type=(\d+)&resource=(\w+)"',
        html,
        re.DOTALL
    )

    # Clean up the hits by stripping whitespace from each captured string
    cleaned_hits = []
    for hit in hits:
        cleaned_hit = tuple(item.strip() for item in hit)  # Strip whitespace from each item
        cleaned_hits.append(cleaned_hit)
    hits = cleaned_hits
        
    offers = []
    for hit in hits:
        offer = {
            "ciudadDestino": hit[0],
            "jugadorAComprar": hit[1],
            "bienesXminuto": int(hit[2]),
            "amountAvailable": int(
                hit[3].replace(",", "").replace(".", "").replace("<", "").replace(" ", "")
            ),
            "tipo": hit[4],
            "precio": int(hit[5]),
            "destinationCityId": hit[6],
            "cityId": hit[7],
            "position": hit[8],
            "type": hit[9],
            "resource": hit[10],
        }

        # Parse CDN Images to material type
        if offer["tipo"] == "//gf2.geo.gfsrv.net/cdn19/c3527b2f694fb882563c04df6d8972.png":
            offer["tipo"] = "wood"
        elif (
            offer["tipo"] == "//gf1.geo.gfsrv.net/cdnc6/94ddfda045a8f5ced3397d791fd064.png"
        ):
            offer["tipo"] = "wine"
        elif (
            offer["tipo"] == "//gf3.geo.gfsrv.net/cdnbf/fc258b990c1a2a36c5aeb9872fc08a.png"
        ):
            offer["tipo"] = "marble"
        elif (
            offer["tipo"] == "//gf2.geo.gfsrv.net/cdn1e/417b4059940b2ae2680c070a197d8c.png"
        ):
            offer["tipo"] = "glass"
        elif (
            offer["tipo"] == "//gf1.geo.gfsrv.net/cdn9b/5578a7dfa3e98124439cca4a387a61.png"
        ):
            offer["tipo"] = "sulfur"
        else:
            continue

        offers.append(offer)
    return offers


def calculateCost(offers, total_amount_to_buy):
    """
    Parameters
    ----------
    offers : list[dict]
    total_amount_to_buy : int
    Returns
    -------
    total_cost : int
    """
    total_cost = 0
    for offer in offers:
        if total_amount_to_buy == 0:
            break
        buy_amount = min(offer["amountAvailable"], total_amount_to_buy)
        total_amount_to_buy -= buy_amount
        total_cost += buy_amount * offer["precio"]
    return total_cost


def chooseCommertialCity(commercial_cities):
    """
    Parameters
    ----------
    commercial_cities : list[dict]

    Returns
    -------
    commercial_city : dict
    """
    print("From which city do you want to buy resources?\n")
    for i, city in enumerate(commercial_cities):
        print("({:d}) {}".format(i + 1, city["name"]))
    selected_city_index = read(min=1, max=len(commercial_cities))
    return commercial_cities[selected_city_index - 1]


def buyResources(session, event, stdin_fd, predetermined_input):
    """
    Parameters
    ----------
    session : ikabot.web.session.Session
    event : multiprocessing.Event
    stdin_fd: int
    predetermined_input : multiprocessing.managers.SyncManager.list
    """
    sys.stdin = os.fdopen(stdin_fd)
    config.predetermined_input = predetermined_input
    try:
        banner()

        # get all the cities with a store
        commercial_cities = getCommercialCities(session)
        if len(commercial_cities) == 0:
            print("There is no store build")
            enter()
            event.set()
            return

        # choose which city to buy from
        if len(commercial_cities) == 1:
            city = commercial_cities[0]
        else:
            city = chooseCommertialCity(commercial_cities)
            banner()

        # choose resource to buy
        resource = chooseResource(session, city)
        banner()

        # get all the offers of the chosen resource from the chosen city
        offers = getOffers(session, city)
        if len(offers) == 0:
            print("There are no offers available.")
            enter()
            event.set()
            return

        # display offers to the user
        total_price = 0
        total_amount = 0
        for offer in offers:
            amount = offer["amountAvailable"]
            price = offer["precio"]
            cost = amount * price
            print("amount:{}".format(addThousandSeparator(amount)))
            print("price :{:d}".format(price))
            print("cost  :{}".format(addThousandSeparator(cost)))
            print("")
            total_price += cost
            total_amount += amount

        # ask how much to buy
        print(
            "Total amount available to purchase: {}, for {}".format(
                addThousandSeparator(total_amount), addThousandSeparator(total_price)
            )
        )
        available = city["freeSpaceForResources"][resource]
        if available < total_amount:
            print(
                "You just can buy {} due to storing capacity".format(
                    addThousandSeparator(available)
                )
            )
            total_amount = available
        print("")
        amount_to_buy = read(
            msg="How much do you want to buy?: ", min=0, max=total_amount
        )
        if amount_to_buy == 0:
            event.set()
            return

        # calculate the total cost
        (gold, __) = getGold(session, city)
        total_cost = calculateCost(offers, amount_to_buy)

        print(
            "\nCurrent gold: {}.\nTotal cost  : {}.\nFinal gold  : {}.".format(
                addThousandSeparator(gold),
                addThousandSeparator(total_cost),
                addThousandSeparator(gold - total_cost),
            )
        )
        print("Proceed? [Y/n]")
        rta = read(values=["y", "Y", "n", "N", ""])
        if rta.lower() == "n":
            event.set()
            return

        print("It will be purchased {}".format(addThousandSeparator(amount_to_buy)))
        enter()
    except KeyboardInterrupt:
        event.set()
        return

    set_child_mode(session)
    event.set()

    info = "\nI will buy {} from {} to {}\n".format(
        addThousandSeparator(amount_to_buy), materials_names[resource], city["cityName"]
    )
    setInfoSignal(session, info)
    try:
        do_it(session, city, offers, amount_to_buy)
    except Exception as e:
        msg = "Error in:\n{}\nCause:\n{}".format(info, traceback.format_exc())
        sendToBot(session, msg)
    finally:
        session.logout()


def buy(session, city, offer, amount_to_buy, ships_available, ship_capacity):
    """
    Parameters
    ----------
    session : ikabot.web.session.Session
    city : dict
    offer : dict
    amount_to_buy : int
    """
    ships = int(math.ceil((Decimal(amount_to_buy) / Decimal(ship_capacity))))
    data_dict = {
        "action": "transportOperations",
        "function": "buyGoodsAtAnotherBranchOffice",
        "cityId": offer["cityId"],
        "destinationCityId": offer["destinationCityId"],
        "oldView": "branchOffice",
        "position": city["pos"],
        "avatar2Name": offer["jugadorAComprar"],
        "city2Name": offer["ciudadDestino"],
        "type": int(offer["type"]),
        "activeTab": "bargain",
        "transportDisplayPrice": 0,
        "premiumTransporter": 0,
        "normalTransportersMax": ships_available,
        "capacity": 5,
        "max_capacity": 5,
        "jetPropulsion": 0,
        "transporters": ships,
        "backgroundView": "city",
        "currentCityId": offer["cityId"],
        "templateView": "takeOffer",
        "currentTab": "bargain",
        "actionRequest": actionRequest,
        "ajax": 1,
    }
    url = "view=takeOffer&destinationCityId={}&oldView=branchOffice&activeTab=bargain&cityId={}&position={}&type={}&resource={}&backgroundView=city&currentCityId={}&templateView=branchOffice&actionRequest={}&ajax=1".format(
        offer["destinationCityId"],
        offer["cityId"],
        offer["position"],
        offer["type"],
        offer["resource"],
        offer["cityId"],
        actionRequest,
    )
    data = session.post(url)
    html = json.loads(data, strict=False)[1][1][1]
    hits = re.findall(r'"tradegood(\d)Price"\s*value="(\d+)', html)
    for hit in hits:
        data_dict["tradegood{}Price".format(hit[0])] = int(hit[1])
        data_dict["cargo_tradegood{}".format(hit[0])] = 0
    hit = re.search(r'"resourcePrice"\s*value="(\d+)', html)
    if hit:
        data_dict["resourcePrice"] = int(hit.group(1))
        data_dict["cargo_resource"] = 0
    resource = offer["resource"]
    if resource == "resource":
        data_dict["cargo_resource"] = amount_to_buy
    else:
        data_dict["cargo_tradegood{}".format(resource)] = amount_to_buy
    session.post(params=data_dict)
    msg = "I buy {} to {} from {}".format(
        addThousandSeparator(amount_to_buy),
        offer["ciudadDestino"],
        offer["jugadorAComprar"],
    )
    sendToBotDebug(session, msg, debugON_buyResources)
    resource_name = offer['tipo']
    session.setStatus(
        f"Bought {addThousandSeparator(amount_to_buy)} {resource_name} for {offer['precio']} gold from {offer['ciudadDestino']} ({offer['jugadorAComprar']}) ---> {city['name']} | {getDateTime()}"
    )


def do_it(session, city, offers, amount_to_buy):
    """
    Parameters
    ----------
    session : ikabot.web.session.Session
    city : dict
    offers : list[dict]
    amount_to_buy : int
    """
    ship_capacity, freighter_capacity = getShipCapacity(session)
    while True:
        for offer in offers:
            if amount_to_buy == 0:
                return
            if offer["amountAvailable"] == 0:
                continue
            ships_available = waitForArrival(session)
            storageCapacity = ships_available * ship_capacity
            buy_amount = min(amount_to_buy, storageCapacity, offer["amountAvailable"])

            amount_to_buy -= buy_amount
            offer["amountAvailable"] -= buy_amount
            buy(session, city, offer, buy_amount, ships_available, ship_capacity)
            # start from the beginning again, so that we always buy from the cheapest offers fisrt
            break



#! /usr/bin/env python3
# -*- coding: utf-8 -*-

import math
import traceback
import re
from decimal import *
from collections import defaultdict

from ikabot.config import *
from ikabot.helpers.botComm import *
from ikabot.helpers.getJson import getCity
from ikabot.helpers.gui import *
from ikabot.helpers.naval import *
from ikabot.helpers.pedirInfo import *
from ikabot.helpers.planRoutes import waitForArrival
from ikabot.helpers.process import set_child_mode
from ikabot.helpers.signals import setInfoSignal
from ikabot.helpers.varios import *
from ikabot.helpers.pedirInfo import getShipCapacity

from ikabot.function.activateMiracle import obtainMiraclesAvailable
from ikabot.function.attackBarbarians import (
    get_barbarians_info,
    get_current_attacks,
    filter_loading,
    filter_traveling,
    wait_until_attack_is_over,
    wait_for_round,
    wait_for_arrival,
    get_barbarians_lv,
    get_units,
    get_unit_data,
    load_troops,
    get_movements,
)


DEFAULT_SCHEMATICS = {
    "WITH_HEPHAESTUS": [
        # TODO Develop the preset here for when Hephaestus is available
    ],
    "WITHOUT_HEPHAESTUS": [
        {
            "level": (1, 9),
            "looting": {"from_float": False, "units": {"302": 1}},
            "needed_units": {"main": {"302": 90, "304": 21}},
            "waves": {
                "1": {
                    "send": [{"from_float": False, "units": {"302": 90, "304": 21}}],
                }
            },
        },
        {
            "level": (10, 19),
            "looting": {"from_float": False, "units": {"305": 12, "308": 50}},
            "needed_units": {"main": {"302": 60, "304": 35, "305": 12, "308": 50}},
            "waves": {
                "1": {
                    "send": [
                        {
                            "from_float": False,
                            "units": {"302": 60, "304": 35, "305": 12, "308": 50},
                        },
                    ],
                },
            },
        },
        {
            "level": (20, 29),
            "looting": {"from_float": False, "units": {"305": 12, "308": 100}},
            "needed_units": {
                "main": {
                    "302": 60,
                    "304": 70,
                    "305": 12,
                    "307": 12,
                    "308": 100,
                    "309": 30,
                    "310": 5,
                }
            },
            "waves": {
                "1": {
                    "send": [
                        {
                            "from_float": False,
                            "units": {
                                "302": 60,
                                "304": 70,
                                "305": 12,
                                "307": 12,
                                "308": 100,
                                "309": 30,
                                "310": 5,
                            },
                        },
                    ],
                },
            },
        },
        {
            "level": (30, 39),
            "looting": {"from_float": False, "units": {"305": 24, "308": 150}},
            "needed_units": {
                "main": {
                    "302": 300,
                    "304": 147,
                    "305": 24,
                    "307": 18,
                    "308": 300,
                    "310": 5,
                    "311": 10,
                }
            },
            "waves": {
                "1": {
                    "send": [
                        {
                            "from_float": False,
                            "units": {
                                "302": 300,
                                "304": 147,
                                "305": 24,
                                "307": 18,
                                "308": 300,
                                "310": 5,
                                "311": 10,
                            },
                        },
                    ],
                },
            },
        },
    ],
}
FIVE_MINUTES = 5 * 60
DEVELOPMENT = False


def autoBarbarians(session, event, stdin_fd, predetermined_input):
    """
    Parameters
    ----------
    session : ikabot.web.session.Session
    event : multiprocessing.Event
    stdin_fd: int
    predetermined_input : multiprocessing.managers.SyncManager.list
    """
    sys.stdin = os.fdopen(stdin_fd)
    config.predetermined_input = predetermined_input
    ship_capacity, freighter_capacity = getShipCapacity(session)
    try:
        banner()
        print(
            
                "{}⚠️ BEWARE - THE BARBARIAN GRIND TO BE CARRIED OUT MORE EFFICIENTLY REQUIRES THE FOLLOWING RESOURCES ⚠️{}\n".format(
                    bcolors.WARNING, bcolors.ENDC
                )
            
        )
        print(
            "- You need to leave at least 3 merchant ships available.",
            "- You leave 2 extra rams available in the city of origin.",
            "- It is not recommended that merchant ships be used during the grind, as maximizing their use increases the efficiency and effectiveness of attacks.",
            sep="\n",
        )
        print(
            
                "\nDo you agree that failure to comply with these rules will result in you losing out on resources? [y/N]"
            
        )
        if read(values=["y", "Y", "n", "N"], default="n") in ["n", "N"]:
            event.set()
            return

        banner()
        island = choose_island(session)
        if island is None:
            event.set()
            return

        banner()
        print("From which city do you want to attack?")
        city = chooseCity(session)
        if city is None:
            event.set()
            return

        has_rams = has_units_in_city(session, city, {"307": 1})
        if has_rams is False:
            print(
                
                    "\nYou do not have 2 or more battering rams in this city, the lack of them may prevent you from collecting all the resources present in the barbarian village, are you sure you want to continue anyway? [y/N]"
                
            )
            if read(values=["y", "Y", "n", "N"], default="n") in ["n", "N"]:
                event.set()
                return

        banner()
        if DEVELOPMENT is True:
            islands = obtainMiraclesAvailable(session)
            hephaestus_max = is_hephaestus_max(islands)
            auto_activate_hephaestus = False
            if hephaestus_max:
                print(
                    
                        "Do you want to keep activating your Hephaestus to maximize the grind? [Y/n]"
                    
                )
                activate_miracle_input = read(values=["y", "Y", "n", "N", ""])
                auto_activate_hephaestus = (
                    True if activate_miracle_input in ("y", "Y") else False
                )

        schematic = DEFAULT_SCHEMATICS["WITHOUT_HEPHAESTUS"]
        if DEVELOPMENT is True:
            banner()
            schematic_option = choose_schematic()
            if schematic_option is None:
                event.set()
                return

            if schematic_option == 1:
                if auto_activate_hephaestus:
                    schematic = DEFAULT_SCHEMATICS["WITH_HEPHAESTUS"]
                pass
            elif schematic_option == 2:
                # TODO do the part where the user can select a custom structure
                pass

        banner()
        success, schematic_informations = get_schematic_information(
            session,
            city,
            schematic,
            ship_capacity,
            is_in_island=True if city["islandId"] == island["id"] else False,
        )
        units_data = schematic_informations["units_data"]
        schematic_units = schematic_informations["schematic_units"]
        main_city_units = schematic_informations["main_city_units"]
        schematic_ships = schematic_informations["schematic_ships"]

        ships_available = waitForArrival(session)
        print(
            "For this sequence of attacks you need to have the following troops:\n"
        )
        print_grid_units(
            schematic_units["total"], main_city_units, schematic_ships, ships_available
        )
        if success is False:
            print(
                
                    "\nYou do not have all the units needed to start this attack sequence, you want to continue executing the attack only as far as possible? [Y/n]"
                
            )
            if read(values=["y", "Y", "n", "N"], default="y") in ["n", "N"]:
                event.set()
                return

        banner()
        need_float_city = check_need_float_city(schematic)
        float_city = None
        if need_float_city is True:
            float_city = choose_float_city(session, island)
            if float_city is None:
                event.set()
                return

    except KeyboardInterrupt:
        event.set()
        return

    set_child_mode(session)
    event.set()

    info = "\nI grinding the barbarians in [{}:{}]\n".format(
        island["x"], island["y"]
    )
    setInfoSignal(session, info)

    try:
        do_it(session, island, city, float_city, schematic, units_data, ship_capacity)
    except Exception as e:
        msg = "Error in:\n{}\nCause:\n{}".format(info, traceback.format_exc())
        sendToBot(session, msg)
    finally:
        session.logout()


def choose_island(session):
    idsIslands = getIslandsIds(session)
    islands = []
    for idIsland in idsIslands:
        html = session.get(island_url + idIsland)
        island = getIsland(html)
        islands.append(island)

    if len(islands) == 1:
        return islands[0]

    islands.sort(key=lambda island: island["id"])

    longest_island_name_length = 0
    for island in islands:
        longest_island_name_length = max(
            len(island["name"]), longest_island_name_length
        )

    def pad(island_name):
        return " " * (longest_island_name_length - len(island_name)) + island_name

    print("In which island do you want to attack the barbarians?")
    print(" 0) Exit")
    for i, island in enumerate(islands):
        num = " " + str(i + 1) if i < 9 else str(i + 1)
        if island["barbarians"]["destroyed"] == 1:
            warn = "(currently destroyed)"
        else:
            warn = ""
        print(
            "{}) [{}:{}] {} ({}) : barbarians lv: {} ({}) {}".format(
                num,
                island["x"],
                island["y"],
                pad(island["name"]),
                materials_names[int(island["tradegood"])][0].upper(),
                island["barbarians"]["level"],
                island["barbarians"]["city"],
                warn,
            )
        )

    index = read(min=0, max=len(islands))
    if index == 0:
        return None
    else:
        return islands[index - 1]


def is_hephaestus_max(islands):
    for island in islands:
        if int(island["wonder"]) == 5 and island["wonderActivationLevel"] == 5:
            return True
    return False


def choose_schematic():
    print("Select what type of attack sequence you will do:")
    print("(0) Exit")
    print("(1) Default")
    print("(2) Custom")
    selected = read(min=0, max=2, digit=True)
    if selected == 0:
        return

    return selected


def check_need_float_city(schematic):
    return any(
        any(
            any(send.get("from_float", False) for send in wave_data.get("send", []))
            for wave_data in item.get("waves", {}).values()
        )
        or item.get("looting", {}).get("from_float", False)
        for item in schematic
    )


def choose_float_city(session, island):
    ids, cities = getIdsOfCities(session)
    inslands_cities = {}
    for city_id in cities:
        city = cities[city_id]

        html = session.get(city_url + str(city["id"]))
        city = getCity(html)

        if island["id"] == city["islandId"]:
            inslands_cities[city_id] = city

    if len(inslands_cities) == 1:
        return next(iter(inslands_cities.values()))

    print("Select the city where the float will be located:\n")
    menu_cities = ""
    longest_city_name_length = 0
    for city_id in inslands_cities:
        length = len(cities[city_id]["name"])
        if length > longest_city_name_length:
            longest_city_name_length = length

    def pad(city_name):
        return " " * (longest_city_name_length - len(city_name) + 2)

    i = 0
    cities_options = {}
    for city_id in ids:
        city = inslands_cities.get(city_id)
        if city is None:
            continue
        i += 1
        cities_options[i] = city_id
        city_name = decodeUnicodeEscape(cities[city_id]["name"])
        menu_cities += "{: >2}: {}{}\n".format(i, city_name, pad(city_name))
    menu_cities = menu_cities[:-1]
    print(menu_cities)
    cities_options_index = read(min=1, max=len(cities_options))
    return inslands_cities.get(cities_options[cities_options_index])


def get_max_schematics_units(schematic):
    max_units = {
        "main": defaultdict(int),
        "float": defaultdict(int),
        "total": defaultdict(int),
    }

    if isinstance(schematic, dict):
        schematic = [schematic]

    for item in schematic:
        needed_units = item.get("needed_units", {})

        for category in ("main", "float"):
            units = needed_units.get(category, {})
            for unit_id, unit_amount in units.items():
                max_units[category][unit_id] = max(
                    max_units[category][unit_id], unit_amount
                )

    for unit_id in set(max_units["main"].keys()).union(max_units["float"].keys()):
        max_units["total"][unit_id] = max_units["main"].get(unit_id, 0) + max_units[
            "float"
        ].get(unit_id, 0)

    return {key: dict(value) for key, value in max_units.items()}


def get_schematic_information(
    session, city, schematic, ship_capacity, float_city=None, is_in_island=False
):
    schematic_units = get_max_schematics_units(schematic)
    main_city_units = get_units(session, city)
    float_city_units = {}
    if float_city is not None:
        float_city_units = get_units(session, float_city)
        pass

    units_data = {}
    for unit_id in schematic_units["total"].keys():
        units_data[unit_id] = get_unit_data(session, city["id"], str(unit_id))

    schematic_ships = 2
    if is_in_island is False:
        schematic_ships += get_amount_ships_schematic(
            schematic_units["total"], units_data, ship_capacity
        )

    def get_success():
        return all(
            [
                all(
                    main_city_units.get(unit_id, {"amount": 0})["amount"] >= amount
                    for unit_id, amount in schematic_units["main"].items()
                ),
                (
                    all(
                        float_city_units.get(unit_id, {"amount": 0})["amount"] >= amount
                        for unit_id, amount in schematic_units["float"].items()
                    )
                    if float_city is not None
                    else True
                ),
            ]
        )

    return get_success(), {
        "units_data": units_data,
        "schematic_units": schematic_units,
        "main_city_units": main_city_units,
        "float_city_units": float_city_units,
        "schematic_ships": schematic_ships,
    }
    # return all(main_city_units.get(unit_id, {'amount': 0})['amount'] >= amount for unit_id, amount in schematic_units.items()),units_data,schematic_units,main_city_units,schematic_ships


def print_grid_units(schematic_units, city_units, schematic_ships, ships_available):
    header = ["Unit", "Required", "Available", "Missing"]
    table = []

    for unit_id, required in schematic_units.items():
        available = city_units.get(unit_id, 0)
        missing = max(0, required - available["amount"])
        table.append([available["name"], required, available["amount"], missing])

    col_widths = [
        max(len(str(row[i])) for row in table + [header]) for i in range(len(header))
    ]
    separator = "+" + "+".join("-" * (w + 2) for w in col_widths) + "+"

    def format_row(row):
        return (
            "| "
            + " | ".join(
                f"{str(cell).ljust(col_widths[i])}" for i, cell in enumerate(row)
            )
            + " |"
        )

    print(separator)
    print(format_row(header))
    print(separator)
    for row in table:
        print(format_row(row))
    print(separator)
    footer = [
        "ships",
        str(schematic_ships),
        str(ships_available),
        max(0, int(schematic_ships) - int(ships_available)),
    ]
    print(format_row(footer))
    print(separator)


def get_barbarians_attack_plan(barbarians_info, schematic):
    barbarians_level = int(barbarians_info["level"])
    selected_scheme = None
    for scheme in schematic:
        if isinstance(scheme["level"], (tuple, list)):
            if barbarians_level in range(scheme["level"][0], (scheme["level"][1] + 1)):
                selected_scheme = scheme
                break
        elif isinstance(scheme["level"], int) and barbarians_level == scheme["level"]:
            selected_scheme = scheme
            break

    if selected_scheme is None:
        return

    return {
        "waves": scheme["waves"],
        "looting": scheme["looting"],
        "needed_units": get_max_schematics_units(scheme),
    }


def get_amount_ships_schematic(schematic_units, units_data, ship_capacity):
    schematic_weight = 0.0
    schematic_ships = 2

    for unit_id, amount in schematic_units.items():
        schematic_weight += amount * units_data[str(unit_id)]["weight"]

    schematic_ships += math.ceil(Decimal(schematic_weight) / Decimal(ship_capacity))
    return schematic_ships


def has_units_in_city(session, city, units):
    if not isinstance(units, (dict)):
        raise TypeError(
            f"The type provided for units is invalid, a type (tuple | dict | list) is expected and its type is {type(units)}"
        )
    city_units = get_units(session, city)

    def has_unit(unit_id, unit_amount):
        unit = city_units.get(str(unit_id))
        if unit is None:
            return False
        return True if unit["amount"] >= unit_amount else False

    return all(
        [has_unit(unit_id, unit_amount) for unit_id, unit_amount in units.items()]
    )


def do_it(session, island, city, float_city, schematic, units_data, ship_capacity):
    attempts = {"ships": 0}
    first_loop = True
    while True:
        if first_loop is False:
            time.sleep(FIVE_MINUTES)
        else:
            first_loop = False
        html = session.get(island_url + island["id"])
        island = getIsland(html)
        babarians_info = get_barbarians_lv(session, island, ship_capacity)
        barbarians_plan = get_barbarians_attack_plan(babarians_info, schematic)
        if barbarians_plan is None:
            sendToBot(
                session,
                
                    "It was not possible to continue the attack on the barbarians because they reached a level that is outside the attack scheme.".format(
                        island["x"], island["y"]
                    )
                ,
            )
            break
        if island["barbarians"]["destroyed"] == 1:
            loot(
                session,
                island,
                city,
                barbarians_plan,
                ship_capacity,
                float_city=float_city,
                units_data=units_data,
            )
            wait_for_looting(session, city, island)
            continue
        ships_available = waitForArrival(session)
        schematic_ships = (
            get_amount_ships_schematic(
                barbarians_plan["needed_units"]["total"], units_data, ship_capacity
            )
            + babarians_info["ships"]
        )
        if schematic_ships > ships_available:
            attempts["ships"] += 1
            session.setStatus(
                "waiting for availability of ({}) boats".format(schematic_ships)
            )
            if attempts["ships"] > 20:
                sendToBot(
                    session,
                    
                        "It was not possible to continue the attack on the barbarians due to the long unavailability of ships."
                    ,
                )
                break
            continue
        if (
            has_units_in_city(session, city, barbarians_plan["needed_units"]["total"])
            is False
        ):
            sendToBot(
                session,
                
                    "It was not possible to continue the attack on the barbarians due to the lack of necessary troops in the city(ies)."
                ,
            )
            break
        sendToBot(
            session,
            
                "Starting attack on barbarians[{}:{}] level ({}).".format(
                    island["x"], island["y"], babarians_info["level"]
                )
            ,
        )
        do_attack(
            session,
            island,
            city,
            barbarians_plan,
            ship_capacity,
            float_city=float_city,
            units_data=units_data,
        )
        wait_until_attack_is_over(session, city, island)
        for attempt_key in attempts.keys():
            attempts[attempt_key] = 0

    babarians_info = get_barbarians_lv(session, island, ship_capacity)
    sendToBot(
        session,
        
            "Ended attack sequence on bariarians[{}:{}].".format(
                island["x"], island["y"], babarians_info["level"]
            )
        ,
    )


def do_attack(session, island, city, schematic, ship_capacity, float_city=None, units_data={}):
    battle_start = None
    babarians_info = None

    i = 0
    for wave_id, wave_data in sorted(schematic["waves"].items()):
        i += 1
        wave_id = int(wave_id)

        major_travel_time = None
        minor_travel_time, minor_travel_city = None, city
        sends_data = split_wave_sends_for_group(wave_id, wave_data)
        float_city_data = None
        if float_city is not None:
            float_city_data = []
            for float_attack_round in sends_data["float_city"]:
                float_attack_data, float_ships_needed, float_travel_time = (
                    get_send_attack_data(
                        session, island, float_city, float_attack_round, units_data, ship_capacity
                    )
                )
                float_city_data.append(
                    {
                        "attack_data": float_attack_data,
                        "travel_time": float_travel_time,
                        "ships_needed": float_ships_needed,
                    }
                )
                if major_travel_time is None or float_travel_time > major_travel_time:
                    major_travel_time = float_travel_time
                if minor_travel_time is None or float_travel_time < minor_travel_time:
                    minor_travel_city = float_city
                    minor_travel_time = float_travel_time

        main_city_data = []
        for main_attack_round in sends_data["main_city"]:
            main_attack_data, main_ships_needed, main_travel_time = (
                get_send_attack_data(
                    session, island, city, main_attack_round, units_data, ship_capacity
                )
            )
            main_city_data.append(
                {
                    "attack_data": main_attack_data,
                    "travel_time": main_travel_time,
                    "ships_needed": main_ships_needed,
                }
            )
            if major_travel_time is None or main_travel_time > major_travel_time:
                major_travel_time = main_travel_time
            if minor_travel_time is None or main_travel_time < minor_travel_time:
                minor_travel_city = city
                minor_travel_time = main_travel_time

        try:
            session.setStatus("Waiting for round (round: {})".format(wave_id))
            wait_for_round(
                session, city, island, major_travel_time, battle_start, wave_id
            )

        except AssertionError:
            # battle ended before expected
            return None

        if babarians_info is None:
            babarians_info = get_barbarians_lv(session, island, ship_capacity)
        if battle_start is None:
            battle_start = time.time() + major_travel_time

        ships_needed = sum([data["ships_needed"] for data in main_city_data])
        if float_city is not None:
            ships_needed += sum([data["ships_needed"] for data in float_city_data])

        assert getTotalShips(session) >= ships_needed, "Insufficient cargo ships!"
        ships_available = waitForArrival(session)
        ships_available -= ships_needed

        looting_wave = True if len(schematic["waves"]) == i else False

        if looting_wave:
            if len(main_city_data) > 0:
                main_city_data[0]["attack_data"]["transporter"] = max(
                    babarians_info["ships"], ships_available
                )
            elif len(float_attack_data) > 0:
                float_attack_data[0]["attack_data"]["transporter"] = max(
                    babarians_info["ships"], ships_available
                )

        if isinstance(main_city_data, list) and len(main_city_data) > 0:
            for data in main_city_data:
                session.post(params=data["attack_data"])

        if isinstance(float_city_data, list) and len(float_city_data) > 0:
            for data in float_city_data:
                session.post(params=data["attack_data"])

        if wave_id == 1:
            wait_for_arrival(session, minor_travel_city, island)


def split_wave_sends_for_group(wave_id, wave_data):
    send_groups = wave_data.get("send", [])

    float_city = []
    main_city = []

    for group_index, group in enumerate(send_groups, start=1):
        group_id = f"w{wave_id}_g{group_index}"

        group_with_id = group.copy()
        group_with_id["id"] = group_id

        if group["from_float"]:
            float_city.append(group_with_id)
        else:
            main_city.append(group_with_id)

    return {"float_city": float_city, "main_city": main_city}


def loot(session, island, city, schematic, ship_capacity, float_city=None, units_data={}):
    session.setStatus("Looting remaining resources")
    barbarian_countdown = None
    while True:
        babarians_info = get_barbarians_lv(session, island, ship_capacity)
        html = session.get(island_url + island["id"])
        island = getIsland(html)
        destroyed = island["barbarians"]["destroyed"] == 1
        resources = babarians_info["resources"]
        if destroyed is False or sum(resources) == 0:
            break

        attacks = get_current_attacks(session, city["id"], island["id"])
        attacks = filter_loading(attacks) + filter_traveling(attacks)
        if len(attacks) > 0:
            break

        ships_available = waitForArrival(session) - 2
        destin_city = (
            city if schematic["looting"]["from_float"] is False else float_city
        )
        attack_data, ships_needed, travel_time = get_send_attack_data(
            session, island, destin_city, schematic, units_data, ship_capacity
        )
        attack_data["transporter"] = min(ships_available, ships_needed)

        # make sure we have time to send the attack
        if barbarian_countdown is None:
            resp = get_barbarians_info(session, island["id"])
            if "barbarianCityCooldownTimer" in resp[2][1]:
                barbarian_countdown = resp[2][1]["barbarianCityCooldownTimer"][
                    "countdown"
                ]["enddate"]

        time_left = barbarian_countdown - time.time()
        send_scatter = False
        if time_left is not None and travel_time > (
            time_left - (travel_time + FIVE_MINUTES)
        ):
            send_scatter = True

        session.post(params=attack_data)

        if send_scatter:
            ram_attack_data, ram_travel_time, _ = get_send_attack_data(
                session, island, destin_city, {"307": 1}, units_data, ship_capacity
            )
            session.post(params=ram_attack_data)
            new_countdown = wait_next_scatter(session, ram_travel_time)
            if new_countdown is not None:
                barbarian_countdown = new_countdown

        time_left = barbarian_countdown - time.time()
        if time_left is not None and travel_time > time_left:
            break
        wait_for_arrival(session, city, island)


def get_send_attack_data(session, island, city, attack_round, units_data, ship_capacity):
    attack_data = {
        "action": "transportOperations",
        "function": "attackBarbarianVillage",
        "actionRequest": actionRequest,
        "islandId": island["id"],
        "destinationCityId": 0,
        "cargo_army_304_upkeep": 3,
        "cargo_army_304": 0,
        "cargo_army_315_upkeep": 1,
        "cargo_army_315": 0,
        "cargo_army_302_upkeep": 4,
        "cargo_army_302": 0,
        "cargo_army_303_upkeep": 3,
        "cargo_army_303": 0,
        "cargo_army_312_upkeep": 15,
        "cargo_army_312": 0,
        "cargo_army_309_upkeep": 45,
        "cargo_army_309": 0,
        "cargo_army_307_upkeep": 15,
        "cargo_army_307": 0,
        "cargo_army_306_upkeep": 25,
        "cargo_army_306": 0,
        "cargo_army_305_upkeep": 30,
        "cargo_army_305": 0,
        "cargo_army_311_upkeep": 20,
        "cargo_army_311": 0,
        "cargo_army_310_upkeep": 10,
        "cargo_army_310": 0,
        "transporter": 0,
        "barbarianVillage": 1,
        "backgroundView": "island",
        "currentIslandId": island["id"],
        "templateView": "plunder",
        "ajax": 1,
    }
    return load_troops(session, city, island, attack_round, units_data, attack_data, ship_capacity)


def get_units_scattered(session, city_id=None):
    if city_id is None:
        city_id = getCurrentCityId(session)
    query = {
        "view": "militaryAdvisor",
        "oldView": "updateGlobalData",
        "cityId": city_id,
        "backgroundView": "city",
        "currentCityId": city_id,
        "templateView": "militaryAdvisor",
        "actionRequest": actionRequest,
        "ajax": 1,
    }

    resp = session.post(params=query)
    resp = json.loads(resp, strict=False)

    scattereds = extract_scattered_units(resp[1][1][1])

    return scattereds


def extract_scattered_units(html):
    scattered_units = []

    start_index = html.find('id="scatteredUnitsSidebar"')
    if start_index == -1:
        return scattered_units
    end_index = html.find("</div>", start_index) + len("</div>")
    html = html[start_index:end_index]

    start_index = html.find('<table class="breakdown_table">')
    if start_index == -1:
        return scattered_units
    end_index = html.find("</table>", start_index) + len("</table>")
    html = html[start_index:end_index]

    date_pattern = r"(\d{1,2}\.\d{1,2}\.\d{4}\s\d{1,2}:\d{1,2}:\d{1,2})"
    number_pattern = r"\b:\s(\d)</td>"

    dates = re.findall(date_pattern, html)
    numbers = re.findall(number_pattern, html)

    num_dates = len(dates)
    troop_counts = numbers[-num_dates:]

    for arrival, troop in zip(dates, troop_counts):
        try:
            timestamp = int(datetime.strptime(arrival, "%d.%m.%Y %H:%M:%S").timestamp())
            scattered_units.append({"arrival_time": timestamp, "amount": int(troop)})
        except ValueError:
            continue  # Ignore poorly formatted input

    return scattered_units


def wait_next_scatter(session, travel_time, city_id=None, old_scatter=None):
    if old_scatter is None:
        old_scatter = get_units_scattered(session, city_id)

        wait_time = travel_time
        wait_time -= time.time()
        wait(wait_time + 5)

        new_scatter = get_units_scattered(session, city_id)
        new_entries = [
            entry
            for entry in new_scatter
            if (entry["arrival_time"], entry["amount"]) not in old_scatter
        ]
        if len(new_entries) == 0:
            return None

        for entry in new_entries:
            if entry["amount"] == 1:
                return entry["arrival_time"]

    pass


def get_current_looting(session, city_id, island_id):
    movements = get_movements(session, city_id)
    curr_looting = []

    for movement in movements:
        if movement["event"]["mission"] != 13:
            continue
        if movement["target"]["islandId"] != int(island_id):
            continue
        if movement["event"]["isReturning"] != 2:
            continue
        if movement["origin"]["cityId"] == -1:
            continue

        curr_looting.append(movement)

    return curr_looting


def wait_for_looting(session, city, island):
    lootings = get_current_looting(session, city["id"], island["id"])
    lootings = filter_loading(lootings) + filter_traveling(lootings, onlyCanAbort=False)
    eventTimes = [looting["eventTime"] for looting in lootings]

    if len(eventTimes) == 0:
        return

    wait_time = max(eventTimes)
    wait_time -= time.time()
    wait(wait_time + 5)

    wait_for_looting(session, city, island)



#! /usr/bin/env python3
# -*- coding: utf-8 -*-

import sys

from ikabot.config import *
from ikabot.helpers.getJson import getCity
from ikabot.helpers.gui import *
from ikabot.helpers.pedirInfo import read


def activateVacationMode(session):
    """
    Parameters
    ----------
    session : ikabot.web.session.Session
    """
    html = session.get()
    city = getCity(html)

    data = {
        "action": "Options",
        "function": "activateVacationMode",
        "actionRequest": actionRequest,
        "backgroundView": "city",
        "currentCityId": city["id"],
        "templateView": "options_umod_confirm",
    }
    session.post(params=data, ignoreExpire=True)


def vacationMode(session, event, stdin_fd, predetermined_input):
    """
    Parameters
    ----------
    session : ikabot.web.session.Session
    event : multiprocessing.Event
    stdin_fd: int
    predetermined_input : multiprocessing.managers.SyncManager.list
    """
    sys.stdin = os.fdopen(stdin_fd)
    config.predetermined_input = predetermined_input
    try:
        banner()
        print("Activate vacation mode? [Y/n]")
        rta = read(values=["y", "Y", "n", "N", ""])
        if rta.lower() == "n":
            event.set()
            return

        activateVacationMode(session)

        print("Vacation mode has been activated.")
        enter()
        event.set()
        clear()
        sys.exit()
    except KeyboardInterrupt:
        event.set()
        return



#! /usr/bin/env python3
# -*- coding: utf-8 -*-

import json
import traceback

from ikabot.config import *
from ikabot.helpers.botComm import *
from ikabot.helpers.getJson import getCity
from ikabot.helpers.gui import *
from ikabot.helpers.pedirInfo import *
from ikabot.helpers.process import set_child_mode
from ikabot.helpers.signals import setInfoSignal
from ikabot.helpers.varios import *



def get_studies(session):
    html = session.get()
    city = getCity(html)
    city_id = city["id"]
    url = "view=researchAdvisor&oldView=updateGlobalData&cityId={0}&backgroundView=city&currentCityId={0}&templateView=researchAdvisor&actionRequest={1}&ajax=1".format(
        city_id, actionRequest
    )
    resp = session.post(url)
    resp = json.loads(resp, strict=False)
    return resp[2][1]


def study(session, studies, num_study):
    html = session.get()
    city = getCity(html)
    city_id = city["id"]
    research_type = studies["js_researchAdvisorChangeResearchType{}".format(num_study)][
        "ajaxrequest"
    ].split("=")[-1]
    url = "action=Advisor&function=doResearch&actionRequest={}&type={}&backgroundView=city&currentCityId={}&templateView=researchAdvisor&ajax=1".format(
        actionRequest, research_type, city_id
    )
    session.post(url)


def experiment(session, experiments, automatic):
    while experiments["qty"] > 0:
        if automatic is True: experiments["qty"] = 999999
        # Validate if material is still there..oterwhise log it and send it via bot
        session.get("view=city&cityId={}".format(experiments["cityID"]), noIndex=True)
        data = session.get("view=updateGlobalData&ajax=1", noIndex=True)
        json_data = json.loads(data, strict=False)
        json_data = json_data[0][1]["headerData"]
        current_glass = int(json_data["currentResources"]["3"])

        if current_glass < 300000:
            if automatic is False:
                sendToBot(
                    session,
                    f"Experiment process ended on {experiments['cityName']}, not enough crystal ({addThousandSeparator(current_glass)})",
                )
                break
            else:
                session.setStatus(
                    f"Experiment skipped in {experiments['cityName']}, not enough crystal ({addThousandSeparator(current_glass)}) @{getDateTime()}"
                )
                time.sleep(241 * 60)# Re-try after 4h1m
                continue

        while True:
            cooldown_url = f'view=academy&cityId={experiments["cityID"]}&position={experiments["pos"]}&backgroundView=city&currentCityId={experiments["cityID"]}&templateView=academy&actionRequest={actionRequest}&ajax=1'
            cooldown_html = session.get(cooldown_url)
            if 'experimentCooldown' in cooldown_html: # Check if a cooldown is still in effect
                session.setStatus(
                    f"Cooldown is still in effect in {experiments['cityName']} @{getDateTime()}",
                )
                time.sleep(300) # Check again after 5 minutes
                continue
            else:
                break

        url = f'action=CityScreen&function=buyResearch&cityId={experiments["cityID"]}&position={experiments["pos"]}&backgroundView=city&currentCityId={experiments["cityID"]}&templateView=academy&actionRequest={actionRequest}&ajax=1'
        session.post(url)

        experiments["qty"] = experiments["qty"] - 1
        if automatic is False:
            session.setStatus(
                f"Experiment done in {experiments['cityName']} @{getDateTime()}, left = {experiments['qty']} time (s)",
            )
        else:
            session.setStatus(
                f"Experiment done in {experiments['cityName']} @{getDateTime()}"
            )

        # Terminate it if no of experiments = 0
        if experiments["qty"] == 0:
            break

        # Every 4h, added 1m extra
        time.sleep(241 * 60)


def investigate(session, event, stdin_fd, predetermined_input):
    """
    Parameters
    ----------
    session : ikabot.web.session.Session
    event : multiprocessing.Event
    stdin_fd: int
    predetermined_input : multiprocessing.managers.SyncManager.list
    """
    sys.stdin = os.fdopen(stdin_fd)
    config.predetermined_input = predetermined_input
    try:
        banner()

        print("\nSelect an option:")
        print("1) Study")
        print("2) Conduct experiment")
        print("3) Conduct automatically")
        option = read(min=1, max=3)

        if option == 1:
            studies = get_studies(session)
            keys = list(studies.keys())
            num_studies = len(
                [
                    key
                    for key in keys
                    if "js_researchAdvisorChangeResearchTypeTxt" in key
                ]
            )

            available = []
            for num_study in range(num_studies):
                if "js_researchAdvisorProgressTxt{}".format(num_study) in studies:
                    available.append(num_study)

            if len(available) == 0:
                print("There are no available studies.")
                enter()
                event.set()
                return

            print("Which one do you wish to study?")
            print("0) None")
            for index, num_study in enumerate(available):
                print(
                    "{:d}) {}".format(
                        index + 1,
                        studies[
                            "js_researchAdvisorNextResearchName{}".format(num_study)
                        ],
                    )
                )
            choice = read(min=0, max=len(available))

            if choice == 0:
                event.set()
                return

            study(session, studies, available[choice - 1])
            print("Done.")
            enter()
            event.set()
        else:
            if option == 3:
                max_experiments = 9999999
                automatic = True
            else:
                automatic = False
            # Experiment
            experiments = {}
            total_glass = 0
            found_academy = -1

            # while (total_glass < 300000 or found_academy < 0):
            banner()
            print("Pick city: ")
            city = chooseCity(session)
            total_glass = int(city["availableResources"][3])

            # Check if enough glass
            if automatic is False:
                if total_glass < 300000:
                    print(
                        f"Not enough glass ({addThousandSeparator(total_glass)}), try another city. Min=300k"
                    )
                    time.sleep(2)
                    enter()
                    event.set()
                    return

            # Search for Academy
            for building in city["position"]:
                if building["building"] == "academy":
                    found_academy = building["position"]

            if found_academy < 0:
                print(f"No academy in this town, pick another one")
                time.sleep(2)
                enter()
                event.set()
                return

            if automatic is False:
                max_experiments = total_glass // 300000
                automatic = False
                banner()
                print(f"How many experiments? Min=1, Max={max_experiments}")
                choice = read(min=1, max=max_experiments)

            if automatic is True:
                choice = 999999

            # Build experiments dict
            experiments["cityID"] = city["id"]
            experiments["cityName"] = city["name"]
            experiments["pos"] = found_academy
            experiments["qty"] = choice

            # Process
            set_child_mode(session)
            event.set()

            if automatic is False:
                info = f"Process: Experiments\n\nWill excecute {choice} times every 4h"
            else:
                info = f"Process: Experiments\n\nWill excecute every 4h"

            try:
                sendToBot(session, info)
                experiment(session, experiments, automatic)
            except Exception as e:
                error_msg = f"Error in:\n{info}\nCause:\n{traceback.format_exc()}"
                sendToBot(session, error_msg)
            finally:
                session.logout()

    except KeyboardInterrupt:
        event.set()
        return



#! /usr/bin/env python3
# -*- coding: utf-8 -*-

import locale
import os

# Version is changed automatically by the release pipeline
IKABOT_VERSION = "7.2.5"


IKABOT_VERSION_TAG = "v" + IKABOT_VERSION




update_msg = ""

isWindows = os.name == "nt"


LOGS_DIRECTORY_FILE = os.getenv("temp") + "/ikabot.log" if isWindows else "/tmp/ikabot.log"
DEFAULT_LOG_LEVEL = 30 # Warning

publicAPIServerDomain = "ikagod.twilightparadox.com"
do_ssl_verify = True
ids_cache = None
cities_cache = None
has_params = False
menu_cities = ""
infoUser = ""
ikaFile = ".ikabot"
city_url = "view=city&cityId="
island_url = "view=island&islandId="
prompt = " >>  "
materials_names = ["Wood", "Wine", "Marble", "Crystal", "Sulfur"]
materials_names_english = ["Wood", "Wine", "Marble", "Crystal", "Sulfur"]
miracle_names_english = [
    "",
    "Hephaestus' Forge",
    "Hades' Holy Grove",
    "Demeter's gardens",
    "Athena's Parthenon",
    "Temple of Hermes",
    "Ares' Stronghold",
    "Temple of Poseidon",
    "Colossus",
]
materials_names_tec = ["wood", "wine", "marble", "glass", "sulfur"]
material_img_hash = [
    "19c3527b2f694fb882563c04df6d8972",
    "c694ddfda045a8f5ced3397d791fd064",
    "bffc258b990c1a2a36c5aeb9872fc08a",
    "1e417b4059940b2ae2680c070a197d8c",
    "9b5578a7dfa3e98124439cca4a387a61",
]
tradegoods_names = [
    "Saw mill",
    "Vineyard",
    "Quarry",
    "Crystal Mine",
    "Sulfur Pit",
]
ConnectionError_wait = 5 * 60
actionRequest = "REQUESTID"
piracyMissionToBuildingLevel = {
    1: 1,
    2: 3,
    3: 5,
    4: 7,
    5: 9,
    6: 11,
    7: 13,
    8: 15,
    9: 17,
}
piracyMissionWaitingTime = {
    1: 150,
    2: 450,
    3: 900,
    4: 1800,
    5: 3600,
    6: 7200,
    7: 14400,
    8: 28800,
    9: 57600,
}
predetermined_input = []




debugON_alertAttacks = False
debugON_alertLowWine = False
debugON_donationBot = False
debugON_searchForIslandSpaces = False
debugON_loginDaily = False
debugON_enviarVino = False
debugON_sendResources = False
debugON_constructionList = False
debugON_buyResources = False
debugON_activateMiracle = False

MAXIMUM_CITY_NAME_LENGTH = 20
SECONDS_IN_HOUR = 60 * 60

# Default values for dynamic settings
enable_CustomPort = False

user_agents = [
    "Mozilla/5.0 (Windows NT 10.0; Win64; x64) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/123.0.0.0 Safari/537.3",
    "Mozilla/5.0 (Windows NT 10.0; Win64; x64) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/122.0.0.0 Safari/537.3",
    "Mozilla/5.0 (Windows NT 10.0; Win64; x64; rv:124.0) Gecko/20100101 Firefox/124.",
    "Mozilla/5.0 (Windows NT 10.0; Win64; x64) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/122.0.0.0 Safari/537.36 Edg/122.0.0.",
    "Mozilla/5.0 (Windows NT 10.0; Win64; x64) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/109.0.0.0 Safari/537.3",
    "Mozilla/5.0 (Windows NT 10.0; Win64; x64) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/117.0.0.0 Safari/537.36 Edg/117.0.2045.4",
    "Mozilla/5.0 (Windows NT 10.0; Win64; x64) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/122.0.0.0 Safari/537.36 OPR/108.0.0.",
    "Mozilla/5.0 (Windows NT 10.0; Win64; x64) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/121.0.0.0 Safari/537.36 Edg/121.0.0.",
    "Mozilla/5.0 (Windows NT 10.0; Win64; x64) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/112.0.0.0 Safari/537.36 Edg/112.0.1722.5",
    "Mozilla/5.0 (Windows NT 10.0; Win64; x64) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/109.0.0.0 Safari/537.36 Edg/109.0.1518.5",
    "Mozilla/5.0 (Windows NT 10.0; Win64; x64) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/121.0.0.0 Safari/537.3",
    "Mozilla/5.0 (X11; Linux x86_64) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/122.0.0.0 Safari/537.3",
    "Mozilla/5.0 (X11; Linux x86_64) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/122.0.0.0 Safari/537.36 OPR/108.0.0.",
    "Mozilla/5.0 (Windows NT 10.0; Win64; x64; rv:123.0) Gecko/20100101 Firefox/123.",
    "Mozilla/5.0 (X11; Linux x86_64) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/115.0.0.0 Safari/537.3",
    "Mozilla/5.0 (Windows NT 6.1; Win64; x64; rv:109.0) Gecko/20100101 Firefox/115.",
    "Mozilla/5.0 (Windows NT 6.1) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/109.0.0.0 Safari/537.36 Edg/109.0.1518.14",
    "Mozilla/5.0 (Windows NT 6.1) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/109.0.0.0 Safari/537.36 Edg/109.0.1518.10",
    "Mozilla/5.0 (Windows NT 10.0; Win64; x64; rv:109.0) Gecko/20100101 Firefox/118.",
    "Mozilla/5.0 (Windows NT 10.0; Win64; x64) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/83.0.4103.106 Safari/537.3",
    "Mozilla/5.0 (Windows NT 10.0; WOW64; Trident/7.0; rv:11.0) like Geck",
    "Mozilla/5.0 (Windows NT 10.0; Win64; x64) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/107.0.0.0 Safari/537.36 Edg/107.0.1418.2",
    "Mozilla/5.0 (Windows NT 10.0; Win64; x64) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/112.0.0.0 Safari/537.36 Edg/112.0.1722.3",
    "Mozilla/5.0 (Windows NT 10.0; Win64; x64) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/123.0.0.0 Safari/537.36 Edg/123.0.0.",
]
